# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy.oem.dell.resources.manager import lifecycle_service


class DellLCServiceTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'lifecycle_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = '1'
        self.lifecycle_service = lifecycle_service.DellLCService(
            self.conn,
            '/redfish/v1/Dell/Managers/iDRAC.Embedded.1/DellLCService')

    def test_is_idrac_ready_true(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "LCStatus": "Ready",
            "RTStatus": "Ready",
            "ServerStatus": "OutOfPOST",
            "Status": "Ready"
            }
        target_uri = ('/redfish/v1/Dell/Managers/iDRAC.Embedded.1'
                      '/DellLCService'
                      '/Actions/DellLCService.GetRemoteServicesAPIStatus')
        idrac_ready_response = self.lifecycle_service.is_idrac_ready()
        self.conn.post.assert_called_once_with(target_uri, data={})
        self.assertTrue(idrac_ready_response)

    def test_is_idrac_ready_false(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.json.return_value = {
            "LCStatus": "NotReady",
            "RTStatus": "NotReady",
            "ServerStatus": "OutOfPOST",
            "Status": "NotReady"
            }
        target_uri = ('/redfish/v1/Dell/Managers/iDRAC.Embedded.1'
                      '/DellLCService'
                      '/Actions/DellLCService.GetRemoteServicesAPIStatus')
        idrac_ready_response = self.lifecycle_service.is_idrac_ready()
        self.conn.post.assert_called_once_with(target_uri, data={})
        self.assertFalse(idrac_ready_response)

    def test_is_realtime_ready_true(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "LCStatus": "Ready",
            "RTStatus": "Ready",
            "ServerStatus": "OutOfPOST",
            "Status": "Ready"
            }

        self.assertTrue(self.lifecycle_service.is_realtime_ready())
        target_uri = ('/redfish/v1/Dell/Managers/iDRAC.Embedded.1'
                      '/DellLCService'
                      '/Actions/DellLCService.GetRemoteServicesAPIStatus')
        self.conn.post.assert_called_once_with(target_uri, data={})

    def test_is_realtime_ready_false(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.json.return_value = {
            "LCStatus": "Ready",
            "RTStatus": "NotReady",
            "ServerStatus": "OutOfPOST",
            "Status": "NotReady"
            }

        self.assertFalse(self.lifecycle_service.is_realtime_ready())
        target_uri = ('/redfish/v1/Dell/Managers/iDRAC.Embedded.1'
                      '/DellLCService'
                      '/Actions/DellLCService.GetRemoteServicesAPIStatus')
        self.conn.post.assert_called_once_with(target_uri, data={})



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import json
from unittest import mock

from oslotest.base import BaseTestCase
import requests

import sushy
from sushy.oem.dell.resources.manager import constants as mgr_cons
from sushy.oem.dell.resources.manager import idrac_card_service as idrac_card
from sushy.oem.dell.resources.manager import job_collection as jc
from sushy.oem.dell.resources.manager import job_service as job
from sushy.oem.dell.resources.manager import lifecycle_service as lifecycle
from sushy.oem.dell.resources.manager import manager as oem_manager
from sushy.resources.manager import manager
from sushy.taskmonitor import TaskMonitor


class ManagerTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'manager.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers = {
            'Location': '/redfish/v1/TaskService/Tasks/JID_905749031119'}
        mock_response.content = None

        self.manager = manager.Manager(self.conn, '/redfish/v1/Managers/BMC',
                                       redfish_version='1.0.2')

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_import_system_configuration_uri(self):
        oem = self.manager.get_oem_extension('Dell')

        self.assertEqual(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration',
            oem.import_system_configuration_uri)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_set_virtual_boot_device_cd(self):
        oem = self.manager.get_oem_extension('Dell')

        oem.set_virtual_boot_device(
            sushy.VIRTUAL_MEDIA_CD, manager=self.manager)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration',
            data={'ShareParameters': {'Target': ['ALL']},
                  'ImportBuffer':
                  '<SystemConfiguration><Component FQDD="iDRAC.Embedded.1">'
                  '<Attribute Name="ServerBoot.1#BootOnce">Enabled'
                  '</Attribute><Attribute Name="ServerBoot.1'
                  '#FirstBootDevice">VCD-DVD</Attribute></Component>'
                  '</SystemConfiguration>'})

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_set_virtual_boot_device_cd_no_manager_passed(self):
        oem = self.manager.get_oem_extension('Dell')

        oem.set_virtual_boot_device(sushy.VIRTUAL_MEDIA_CD)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration',
            data={'ShareParameters': {'Target': ['ALL']},
                  'ImportBuffer':
                  '<SystemConfiguration><Component FQDD="iDRAC.Embedded.1">'
                  '<Attribute Name="ServerBoot.1#BootOnce">Enabled'
                  '</Attribute><Attribute Name="ServerBoot.1'
                  '#FirstBootDevice">VCD-DVD</Attribute></Component>'
                  '</SystemConfiguration>'})

    @mock.patch('time.sleep', autospec=True)
    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_set_virtual_boot_device_cd_running_exc(self, mock_sleep):
        oem = self.manager.get_oem_extension('Dell')

        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'error_running_job.json') as f:
            response_obj = json.load(f)

        response1 = mock.MagicMock(spec=requests.Response)
        response1.status_code = http_client.BAD_REQUEST
        response1.json.return_value = response_obj
        response1.code = "IDRAC.2.8.LC068"

        response2 = mock.MagicMock(spec=requests.Response)
        response2.status_code = http_client.OK

        self.conn.post.side_effect = [sushy.exceptions.HTTPError(
            method='POST', url=self.manager.path, response=response1),
            response2]

        oem.set_virtual_boot_device(
            sushy.VIRTUAL_MEDIA_CD, manager=self.manager)

        self.conn.post.assert_called_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration',
            data={'ShareParameters': {'Target': ['ALL']},
                  'ImportBuffer':
                  '<SystemConfiguration><Component FQDD="iDRAC.Embedded.1">'
                  '<Attribute Name="ServerBoot.1#BootOnce">Enabled'
                  '</Attribute><Attribute Name="ServerBoot.1'
                  '#FirstBootDevice">VCD-DVD</Attribute></Component>'
                  '</SystemConfiguration>'})

    @mock.patch('sushy.oem.dell.utils.reboot_system', autospec=True)
    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_set_virtual_boot_device_cd_pending_exc(self, mock_reboot):
        oem = self.manager.get_oem_extension('Dell')

        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'error_pending_job.json') as f:
            response_obj = json.load(f)

        response1 = mock.MagicMock(spec=requests.Response)
        response1.status_code = http_client.BAD_REQUEST
        response1.json.return_value = response_obj
        response1.code = "IDRAC.2.8.LC068"

        response2 = mock.MagicMock(spec=requests.Response)
        response2.status_code = http_client.OK

        self.conn.post.side_effect = [sushy.exceptions.HTTPError(
            method='POST', url=self.manager.path, response=response1),
            response2]

        oem.set_virtual_boot_device(
            sushy.VIRTUAL_MEDIA_CD, manager=self.manager)

        self.conn.post.assert_called_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration',
            data={'ShareParameters': {'Target': ['ALL']},
                  'ImportBuffer':
                  '<SystemConfiguration><Component FQDD="iDRAC.Embedded.1">'
                  '<Attribute Name="ServerBoot.1#BootOnce">Enabled'
                  '</Attribute><Attribute Name="ServerBoot.1'
                  '#FirstBootDevice">VCD-DVD</Attribute></Component>'
                  '</SystemConfiguration>'})

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_set_virtual_boot_device_cd_other_exc(self):
        oem = self.manager.get_oem_extension('Dell')

        response = mock.MagicMock(spec=requests.Response)
        response.status_code = http_client.FORBIDDEN

        self.conn.post.side_effect = sushy.exceptions.HTTPError(
            method='POST', url=self.manager.path, response=response)

        self.assertRaises(sushy.exceptions.HTTPError,
                          oem.set_virtual_boot_device,
                          sushy.VIRTUAL_MEDIA_CD,
                          manager=self.manager)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_allowed_export_target_values(self):
        oem = self.manager.get_oem_extension('Dell')
        expected_values = {mgr_cons.ExportTarget.IDRAC,
                           mgr_cons.ExportTarget.RAID,
                           mgr_cons.ExportTarget.ALL,
                           mgr_cons.ExportTarget.BIOS,
                           mgr_cons.ExportTarget.NIC}
        allowed_values = oem.get_allowed_export_target_values()
        self.assertEqual(expected_values, allowed_values)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_allowed_export_target_values_missing(self):
        oem = self.manager.get_oem_extension('Dell')
        export_action = ('OemManager.v1_0_0'
                         '#OemManager.ExportSystemConfiguration')
        oem.json['Actions']['Oem'][export_action]['ShareParameters'].pop(
            'Target@Redfish.AllowableValues')
        oem.refresh()
        oem = self.manager.get_oem_extension('Dell')
        expected_values = {mgr_cons.ExportTarget.IDRAC,
                           mgr_cons.ExportTarget.RAID,
                           mgr_cons.ExportTarget.ALL,
                           mgr_cons.ExportTarget.BIOS,
                           mgr_cons.ExportTarget.NIC}
        allowed_values = oem.get_allowed_export_target_values()
        self.assertEqual(expected_values, allowed_values)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_export_system_configuration_uri(self):
        oem = self.manager.get_oem_extension('Dell')

        self.assertEqual(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ExportSystemConfiguration',
            oem.export_system_configuration_uri)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test__export_system_configuration(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration(
            target=mgr_cons.ExportTarget.ALL)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ExportSystemConfiguration', data={'ShareParameters':
                                                {'Target': ['ALL']},
                                                'ExportFormat': 'JSON',
                                                'ExportUse': 'Default',
                                                'IncludeInExport': 'Default'})

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test__export_system_configuration_nondefault(self):
        oem = self.manager.get_oem_extension('Dell')
        include_in_export = mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES
        oem._export_system_configuration(
            target=mgr_cons.ExportTarget.RAID,
            export_use=mgr_cons.ExportUse.CLONE,
            include_in_export=include_in_export)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ExportSystemConfiguration', data={'ShareParameters':
                                                {'Target': ['RAID']},
                                                'ExportFormat': 'JSON',
                                                'ExportUse': 'Clone',
                                                'IncludeInExport':
                                                    'IncludeReadOnly,Include'
                                                    'PasswordHashValues'})

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test__export_system_configuration_invalid_target(self):
        oem = self.manager.get_oem_extension('Dell')
        target = "xyz"
        self.assertRaises(sushy.exceptions.InvalidParameterValueError,
                          oem._export_system_configuration, target)

    def test__export_system_configuration_invalid_export_use(self):
        oem = self.manager.get_oem_extension('Dell')
        self.assertRaises(sushy.exceptions.InvalidParameterValueError,
                          oem._export_system_configuration,
                          mgr_cons.ExportTarget.RAID,
                          export_use="ABC")

    def test__export_system_configuration_invalid_include_in_export(self):
        oem = self.manager.get_oem_extension('Dell')
        self.assertRaises(sushy.exceptions.InvalidParameterValueError,
                          oem._export_system_configuration,
                          mgr_cons.ExportTarget.RAID,
                          include_in_export="ABC")

    def test__export_system_configuration_invalid_include_in_export_part(self):
        oem = self.manager.get_oem_extension('Dell')
        export_action = ('OemManager.v1_0_0'
                         '#OemManager.ExportSystemConfiguration')
        # Remove `IncludePasswordHashValues` from allowed values
        oem.json['Actions']['Oem'][export_action]['IncludeInExport@Redfish.'
                                                  'AllowableValues'] =\
            ['Default', 'IncludeReadOnly']
        oem.refresh()
        oem = self.manager.get_oem_extension('Dell')
        include_in_export = mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES
        self.assertRaises(sushy.exceptions.InvalidParameterValueError,
                          oem._export_system_configuration,
                          mgr_cons.ExportTarget.RAID,
                          include_in_export=include_in_export)

    def test__export_system_configuration_include_in_export_legacy(
            self):
        oem = self.manager.get_oem_extension('Dell')
        export_action = ('OemManager.v1_0_0'
                         '#OemManager.ExportSystemConfiguration')
        # Add `IncludeReadOnly,IncludePasswordHashValues` to allowed values
        oem.json['Actions']['Oem'][export_action]['IncludeInExport@Redfish.'
                                                  'AllowableValues'] =\
            ['Default', 'IncludeReadOnly', 'IncludePasswordHashValues',
             'IncludeReadOnly,IncludePasswordHashValues']
        oem.refresh()

        include_in_export = mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES
        oem._export_system_configuration(
            target=mgr_cons.ExportTarget.RAID,
            export_use=mgr_cons.ExportUse.CLONE,
            include_in_export=include_in_export)

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ExportSystemConfiguration', data={'ShareParameters':
                                                {'Target': ['RAID']},
                                                'ExportFormat': 'JSON',
                                                'ExportUse': 'Clone',
                                                'IncludeInExport':
                                                    'IncludeReadOnly,Include'
                                                    'PasswordHashValues'})

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_allowed_export_use_values(self):
        oem = self.manager.get_oem_extension('Dell')
        expected_values = {mgr_cons.ExportUse.DEFAULT,
                           mgr_cons.ExportUse.CLONE,
                           mgr_cons.ExportUse.REPLACE}
        allowed_values = oem.get_allowed_export_use_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)

    @mock.patch.object(oem_manager, 'LOG', autospec=True)
    def test_get_allowed_export_use_values_missing(self, mock_log):
        oem = self.manager.get_oem_extension('Dell')
        export_action = ('OemManager.v1_0_0'
                         '#OemManager.ExportSystemConfiguration')
        oem.json['Actions']['Oem'][export_action].pop(
            'ExportUse@Redfish.AllowableValues')
        oem.refresh()
        expected_values = {mgr_cons.ExportUse.DEFAULT,
                           mgr_cons.ExportUse.CLONE,
                           mgr_cons.ExportUse.REPLACE}
        allowed_values = oem.get_allowed_export_use_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)
        mock_log.warning.assert_called_once()

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_allowed_include_in_export_values(self):
        oem = self.manager.get_oem_extension('Dell')
        expected_values = {mgr_cons.IncludeInExport.DEFAULT,
                           mgr_cons.IncludeInExport.READ_ONLY,
                           mgr_cons.IncludeInExport.PASSWORD_HASHES}
        allowed_values = oem.get_allowed_include_in_export_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)

    @mock.patch.object(oem_manager, 'LOG', autospec=True)
    def test_get_allowed_include_in_export_values_missing(self, mock_log):
        oem = self.manager.get_oem_extension('Dell')
        export_action = ('OemManager.v1_0_0'
                         '#OemManager.ExportSystemConfiguration')
        oem.json['Actions']['Oem'][export_action].pop(
            'IncludeInExport@Redfish.AllowableValues')
        oem.refresh()
        expected_values = {mgr_cons.IncludeInExport.DEFAULT,
                           mgr_cons.IncludeInExport.READ_ONLY,
                           mgr_cons.IncludeInExport.PASSWORD_HASHES,
                           mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES}
        allowed_values = oem.get_allowed_include_in_export_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)
        mock_log.warning.assert_called_once()

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_export_system_configuration(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration = mock.Mock()
        mock_response = mock.Mock()
        oem._export_system_configuration.return_value = mock_response

        response = oem.export_system_configuration()

        self.assertEqual(mock_response, response)
        include_in_export = mgr_cons.IncludeInExport.READ_ONLY_PASSWORD_HASHES
        oem._export_system_configuration.assert_called_once_with(
            mgr_cons.ExportTarget.ALL,
            export_use=mgr_cons.ExportUse.CLONE,
            include_in_export=include_in_export)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_export_system_configuration_destructive_fields(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'export_configuration_idrac.json') as f:
            mock_response = oem._export_system_configuration.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        response = oem.export_system_configuration(
            include_destructive_fields=False)

        response_json = json.loads(response._content)
        # From 40 items in test data 16 should be removed
        self.assertEqual(24, len(response_json['SystemConfiguration']
                                 ['Components'][0]['Attributes']))
        include_in_export = mgr_cons.INCLUDE_EXPORT_READ_ONLY_PASSWORD_HASHES
        oem._export_system_configuration.assert_called_once_with(
            mgr_cons.EXPORT_TARGET_ALL,
            export_use=mgr_cons.EXPORT_USE_CLONE,
            include_in_export=include_in_export)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_pxe_port_macs_bios(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'export_configuration_nic_bios.json') as f:
            mock_response = oem._export_system_configuration.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        ethernet_interfaces_mac = {'NIC.Integrated.1-4-1': '68:05:CA:AF:AA:C9',
                                   'NIC.Slot.7-2-1': '3C:FD:FE:CD:67:31',
                                   'NIC.Slot.7-1-1': '3C:FD:FE:CD:67:30',
                                   'NIC.Integrated.1-2-1': '68:05:CA:AF:AA:C7',
                                   'NIC.Integrated.1-3-1': '68:05:CA:AF:AA:C8',
                                   'NIC.Integrated.1-1-1': '68:05:CA:AF:AA:C6'}

        self.assertEqual(["68:05:CA:AF:AA:C8"],
                         oem.get_pxe_port_macs_bios(ethernet_interfaces_mac))

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_pxe_port_macs_bios_invalid_system_config_tag(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration = mock.Mock()
        mock_response = oem._export_system_configuration.return_value
        mock_response.json.return_value = {'Model': 'PowerEdge R7525'}
        mock_response.status_code = 200
        ethernet_interfaces_mac = {'NIC.Integrated.1-4-1': '68:05:CA:AF:AA:C9',
                                   'NIC.Slot.7-2-1': '3C:FD:FE:CD:67:31',
                                   'NIC.Slot.7-1-1': '3C:FD:FE:CD:67:30',
                                   'NIC.Integrated.1-2-1': '68:05:CA:AF:AA:C7',
                                   'NIC.Integrated.1-3-1': '68:05:CA:AF:AA:C8',
                                   'NIC.Integrated.1-1-1': '68:05:CA:AF:AA:C6'}

        self.assertRaises(sushy.exceptions.ExtensionError,
                          oem.get_pxe_port_macs_bios, ethernet_interfaces_mac)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_get_pxe_port_macs_bios_invalid_response(self):
        oem = self.manager.get_oem_extension('Dell')
        oem._export_system_configuration = mock.Mock()
        mock_response = oem._export_system_configuration.return_value
        mock_response.status_code = 204
        ethernet_interfaces_mac = {'NIC.Integrated.1-4-1': '68:05:CA:AF:AA:C9',
                                   'NIC.Slot.7-2-1': '3C:FD:FE:CD:67:31',
                                   'NIC.Slot.7-1-1': '3C:FD:FE:CD:67:30',
                                   'NIC.Integrated.1-2-1': '68:05:CA:AF:AA:C7',
                                   'NIC.Integrated.1-3-1': '68:05:CA:AF:AA:C8',
                                   'NIC.Integrated.1-1-1': '68:05:CA:AF:AA:C6'}

        self.assertRaises(sushy.exceptions.ExtensionError,
                          oem.get_pxe_port_macs_bios, ethernet_interfaces_mac)

    def test_idrac_card_service(self):
        oem = self.manager.get_oem_extension('Dell')
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'idrac_card_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        idrac_card_service = oem.idrac_card_service
        self.assertEqual(
            '/redfish/v1/Dell/Managers/iDRAC.Embedded.1/DelliDRACCardService',
            idrac_card_service.path)
        self.assertIsInstance(idrac_card_service,
                              idrac_card.DelliDRACCardService)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_lifecycle_service(self):
        oem = self.manager.get_oem_extension('Dell')
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'lifecycle_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        lifecycle_service = oem.lifecycle_service
        self.assertEqual(
            '/redfish/v1/Dell/Managers/iDRAC.Embedded.1/DellLCService',
            lifecycle_service.path)
        self.assertIsInstance(lifecycle_service,
                              lifecycle.DellLCService)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_job_service(self):
        oem = self.manager.get_oem_extension('Dell')
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'job_service.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        job_service = oem.job_service
        self.assertEqual(
            '/redfish/v1/Dell/Managers/iDRAC.Embedded.1/DellJobService',
            job_service.path)
        self.assertIsInstance(job_service,
                              job.DellJobService)

    @mock.patch('sushy.resources.oem.common._global_extn_mgrs_by_resource', {})
    def test_job_collection(self):
        oem = self.manager.get_oem_extension('Dell')
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'job_collection_expanded.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200
        job_collection = oem.job_collection
        self.assertEqual(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Jobs',
            job_collection.path)
        self.assertIsInstance(job_collection,
                              jc.DellJobCollection)

    def test_get_allowed_import_shutdown_type_values(self):
        oem = self.manager.get_oem_extension('Dell')
        expected_values = {mgr_cons.ShutdownType.GRACEFUL,
                           mgr_cons.ShutdownType.FORCED,
                           mgr_cons.ShutdownType.NO_REBOOT}
        allowed_values = oem.get_allowed_import_shutdown_type_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)

    @mock.patch.object(oem_manager, 'LOG', autospec=True)
    def test_get_allowed_import_shutdown_type_values_missing(self, mock_log):
        oem = self.manager.get_oem_extension('Dell')
        import_action = ('OemManager.v1_0_0'
                         '#OemManager.ImportSystemConfiguration')
        oem.json['Actions']['Oem'][import_action].pop(
            'ShutdownType@Redfish.AllowableValues')
        oem.refresh()
        expected_values = {mgr_cons.ShutdownType.GRACEFUL,
                           mgr_cons.ShutdownType.FORCED,
                           mgr_cons.ShutdownType.NO_REBOOT}
        allowed_values = oem.get_allowed_import_shutdown_type_values()
        self.assertIsInstance(allowed_values, set)
        self.assertEqual(expected_values, allowed_values)
        mock_log.warning.assert_called_once()

    def test_import_system_configuration(self):
        oem = self.manager.get_oem_extension('Dell')

        result = oem.import_system_configuration('{"key": "value"}')

        self.conn.post.assert_called_once_with(
            '/redfish/v1/Managers/iDRAC.Embedded.1/Actions/Oem/EID_674_Manager'
            '.ImportSystemConfiguration', data={'ShareParameters':
                                                {'Target': ['ALL']},
                                                'ImportBuffer':
                                                '{"key": "value"}',
                                                'ShutdownType': 'NoReboot'})
        self.assertIsInstance(result, TaskMonitor)
        self.assertEqual('/redfish/v1/TaskService/Tasks/JID_905749031119',
                         result.task_monitor_uri)

    def test_reset_idrac_with_wait_true(self):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager.idrac_card_service.reset_idrac = mock.Mock()
        oem_manager._conn._url = "https://1.2.3.4"
        oem_manager._wait_for_idrac = mock.Mock()
        oem_manager._wait_until_idrac_is_ready = mock.Mock()
        oem_manager.reset_idrac(wait=True)
        oem_manager.idrac_card_service.reset_idrac.assert_called()
        oem_manager._wait_for_idrac.assert_called_with('1.2.3.4', 60)
        oem_manager._wait_until_idrac_is_ready.assert_called_with(
            '1.2.3.4', 96, 10)

    def test_reset_idrac_with_wait_false(self):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager.idrac_card_service.reset_idrac = mock.Mock()
        oem_manager._wait_for_idrac = mock.Mock()
        oem_manager._wait_until_idrac_is_ready = mock.Mock()
        oem_manager.reset_idrac(wait=False)
        oem_manager.idrac_card_service.reset_idrac.assert_called()
        oem_manager._wait_for_idrac.assert_not_called()
        oem_manager._wait_until_idrac_is_ready.assert_not_called()

    def test__wait_until_idrac_is_ready(self):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager.lifecycle_service.is_idrac_ready = mock.Mock()
        oem_manager.lifecycle_service.is_idrac_ready.side_effect = \
            [False, True]
        oem_manager._wait_until_idrac_is_ready('1.2.3.4', 96, 10)
        oem_manager.lifecycle_service.is_idrac_ready.assert_called_with()

    @mock.patch('time.sleep', autospec=True)
    def test__wait_until_idrac_is_ready_with_timeout(self, mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager.lifecycle_service.is_idrac_ready = mock.Mock()
        oem_manager.lifecycle_service.is_idrac_ready.return_value = False
        self.assertRaises(sushy.exceptions.ExtensionError,
                          oem_manager._wait_until_idrac_is_ready,
                          '1.2.3.4', 96, 10)

    @mock.patch('time.sleep', autospec=True)
    def test__wait_for_idrac_with_state_reached(self, mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager._wait_for_idrac_state = mock.Mock()
        oem_manager._wait_for_idrac_state.return_value = True
        oem_manager._wait_for_idrac('1.2.3.4', 30)
        oem_manager._wait_for_idrac_state.assert_called_with(
            '1.2.3.4', alive=True, ping_count=3, retries=24)
        oem_manager._wait_for_idrac_state.assert_any_call(
            '1.2.3.4', alive=False, ping_count=2, retries=24)
        self.assertEqual(2, oem_manager._wait_for_idrac_state.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test__wait_for_idrac_with_first_state_not_reached(self,
                                                          mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager._wait_for_idrac_state = mock.Mock()
        oem_manager._wait_for_idrac_state.return_value = False
        self.assertRaises(sushy.exceptions.ExtensionError,
                          oem_manager._wait_for_idrac, '1.2.3.4', 30)

    @mock.patch('time.sleep', autospec=True)
    def test__wait_for_idrac_with_second_state_not_reached(self,
                                                           mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager._wait_for_idrac_state = mock.Mock()
        oem_manager._wait_for_idrac_state.side_effect = [True, False]
        self.assertRaises(sushy.exceptions.ExtensionError,
                          oem_manager._wait_for_idrac, '1.2.3.4', 30)

    @mock.patch('time.sleep', autospec=True)
    def test__wait_for_idrac_state_with_pingable(self, mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager._ping_host = mock.Mock()
        oem_manager._ping_host.return_value = True
        response = oem_manager._wait_for_idrac_state('1.2.3.4')
        self.assertEqual(True, response)
        self.assertEqual(3, oem_manager._ping_host.call_count)

    @mock.patch('time.sleep', autospec=True)
    def test__wait_for_idrac_state_without_pingable(self, mock_time_sleep):
        oem_manager = self.manager.get_oem_extension('Dell')
        oem_manager._ping_host = mock.Mock()
        oem_manager._ping_host.return_value = False
        response = oem_manager._wait_for_idrac_state('1.2.3.4')
        self.assertEqual(False, response)
        self.assertEqual(24, oem_manager._ping_host.call_count)

    @mock.patch('subprocess.call', autospec=True)
    def test__ping_host_alive(self, mock_call):
        oem_manager = self.manager.get_oem_extension('Dell')
        mock_call.return_value = 0

        result = oem_manager._ping_host('1.2.3.4')

        self.assertTrue(result)
        mock_call.assert_called_with(["ping", "-c", "1", '1.2.3.4'])

    @mock.patch('subprocess.call', autospec=True)
    def test__ping_host_not_alive(self, mock_call):
        oem_manager = self.manager.get_oem_extension('Dell')
        mock_call.return_value = 1

        result = oem_manager._ping_host('1.2.3.4')

        self.assertFalse(result)
        mock_call.assert_called_with(["ping", "-c", "1", '1.2.3.4'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values comes from the Redfish System json-schema 1.0.0:
# http://redfish.dmtf.org/schemas/v1/Manager.v1_0_0.json#/definitions/Manager  # noqa

import enum

from sushy.resources import constants as res_cons

# Manager Reset action constants

RESET_MANAGER_GRACEFUL_RESTART = res_cons.ResetType.GRACEFUL_RESTART
"""Perform a graceful shutdown followed by a restart of the system"""

RESET_MANAGER_FORCE_RESTART = res_cons.ResetType.FORCE_RESTART
"""Perform an immediate (non-graceful) shutdown, followed by a restart"""


class ManagerType(enum.Enum):
    """Manager Type constants"""

    MANAGEMENT_CONTROLLER = 'ManagementController'
    """A controller that primarily monitors or manages the operation of a
    device or system."""

    ENCLOSURE_MANAGER = 'EnclosureManager'
    """A controller that provides management functions for a chassis or
    group of devices or systems."""

    BMC = 'BMC'
    """A controller that provides management functions for a single computer
    system."""

    RACK_MANAGER = 'RackManager'
    """A controller that provides management functions for a whole or part
    of a rack."""

    AUXILIARY_CONTROLLER = 'AuxiliaryController'
    """A controller that provides management functions for a particular
    subsystem or group of devices."""

    SERVICE = 'Service'
    """A software-based service that provides management functions."""


# Backward compatibility
MANAGER_TYPE_MANAGEMENT_CONTROLLER = ManagerType.MANAGEMENT_CONTROLLER
MANAGER_TYPE_ENCLOSURE_MANAGER = ManagerType.ENCLOSURE_MANAGER
MANAGER_TYPE_BMC = ManagerType.BMC
MANAGER_TYPE_RACK_MANAGER = ManagerType.RACK_MANAGER
MANAGER_TYPE_AUXILIARY_CONTROLLER = ManagerType.AUXILIARY_CONTROLLER


class GraphicalConnectType(enum.Enum):
    """Graphical Console constants"""

    KVMIP = 'KVMIP'
    """The controller supports a graphical console connection through a KVM-
    IP (redirection of Keyboard, Video, Mouse over IP) protocol."""

    OEM = 'Oem'
    """The controller supports a graphical console connection through an
    OEM-specific protocol."""


# Backward compatibility
GRAPHICAL_CONSOLE_KVMIP = GraphicalConnectType.KVMIP
GRAPHICAL_CONSOLE_OEM = GraphicalConnectType.OEM


class SerialConnectType(enum.Enum):
    """Serial Console constants"""

    SSH = 'SSH'
    """The controller supports a serial console connection through the SSH
    protocol."""

    TELNET = 'Telnet'
    """The controller supports a serial console connection through the
    Telnet protocol."""

    IPMI = 'IPMI'
    """The controller supports a serial console connection through the IPMI
    Serial Over LAN (SOL) protocol."""

    OEM = 'Oem'
    """The controller supports a serial console connection through an OEM-
    specific protocol."""


# Backward compatibility
SERIAL_CONSOLE_SSH = SerialConnectType.SSH
SERIAL_CONSOLE_TELNET = SerialConnectType.TELNET
SERIAL_CONSOLE_IPMI = SerialConnectType.IPMI
SERIAL_CONSOLE_OEM = SerialConnectType.OEM


class CommandConnectType(enum.Enum):
    """Command Shell constants"""

    SSH = 'SSH'
    """The controller supports a command shell connection through the SSH
    protocol."""

    TELNET = 'Telnet'
    """The controller supports a command shell connection through the Telnet
    protocol."""

    IPMI = 'IPMI'
    """The controller supports a command shell connection through the IPMI
    Serial Over LAN (SOL) protocol."""

    OEM = 'Oem'
    """The controller supports a command shell connection through an OEM-
    specific protocol."""


# Backward compatibility
COMMAND_SHELL_SSH = CommandConnectType.SSH
COMMAND_SHELL_TELNET = CommandConnectType.TELNET
COMMAND_SHELL_IPMI = CommandConnectType.IPMI
COMMAND_SHELL_OEM = CommandConnectType.OEM


class VirtualMediaType(enum.Enum):
    """Supported Virtual Media Type constants"""

    CD = 'CD'
    """A CD-ROM format (ISO) image."""

    FLOPPY = 'Floppy'
    """A floppy disk image."""

    USB_STICK = 'USBStick'
    """An emulation of a USB storage device."""

    DVD = 'DVD'
    """A DVD-ROM format image."""


# Backward compatibility
VIRTUAL_MEDIA_CD = VirtualMediaType.CD
VIRTUAL_MEDIA_DVD = VirtualMediaType.DVD
VIRTUAL_MEDIA_FLOPPY = VirtualMediaType.FLOPPY
VIRTUAL_MEDIA_USBSTICK = VirtualMediaType.USB_STICK


class ConnectedVia(enum.Enum):
    """Connected Via constants"""

    NOT_CONNECTED = 'NotConnected'
    """No current connection."""

    URI = 'URI'
    """Connected to a URI location."""

    APPLET = 'Applet'
    """Connected to a client application."""

    OEM = 'Oem'
    """Connected through an OEM-defined method."""


# Backward compatibility
CONNECTED_VIA_NOT_CONNECTED = ConnectedVia.NOT_CONNECTED
CONNECTED_VIA_URI = ConnectedVia.URI
CONNECTED_VIA_APPLET = ConnectedVia.APPLET
CONNECTED_VIA_OEM = ConnectedVia.OEM


class TransferMethod(enum.Enum):
    """Transfer methods"""

    STREAM = 'Stream'
    """Stream image file data from the source URI."""

    UPLOAD = 'Upload'
    """Upload the entire image file from the source URI to the service."""



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values comes from the Redfish Chassis json-schema:
# https://redfish.dmtf.org/schemas/v1/Chassis.v1_17_0.json

import enum


class ChassisType(enum.Enum):
    """Chassis Types constants"""

    RACK = 'Rack'
    """An equipment rack, typically a 19-inch wide freestanding unit."""

    BLADE = 'Blade'
    """An enclosed or semi-enclosed, typically vertically-oriented, system
    chassis that must be plugged into a multi-system chassis to function
    normally."""

    ENCLOSURE = 'Enclosure'
    """A generic term for a chassis that does not fit any other description."""

    STAND_ALONE = 'StandAlone'
    """A single, free-standing system, commonly called a tower or desktop
    chassis."""

    RACK_MOUNT = 'RackMount'
    """A single-system chassis designed specifically for mounting in an
    equipment rack."""

    CARD = 'Card'
    """A loose device or circuit board intended to be installed in a system
    or other enclosure."""

    CARTRIDGE = 'Cartridge'
    """A small self-contained system intended to be plugged into a multi-
    system chassis."""

    ROW = 'Row'
    """A collection of equipment racks."""

    POD = 'Pod'
    """A collection of equipment racks in a large, likely transportable,
    container."""

    EXPANSION = 'Expansion'
    """A chassis that expands the capabilities or capacity of another
    chassis."""

    SIDECAR = 'Sidecar'
    """A chassis that mates mechanically with another chassis to expand its
    capabilities or capacity."""

    ZONE = 'Zone'
    """A logical division or portion of a physical chassis that contains
    multiple devices or systems that cannot be physically separated."""

    SLED = 'Sled'
    """An enclosed or semi-enclosed, system chassis that must be plugged
    into a multi-system chassis to function normally similar to a blade
    type chassis."""

    SHELF = 'Shelf'
    """An enclosed or semi-enclosed, typically horizontally-oriented, system
    chassis that must be plugged into a multi-system chassis to function
    normally."""

    DRAWER = 'Drawer'
    """An enclosed or semi-enclosed, typically horizontally-oriented, system
    chassis that can be slid into a multi-system chassis."""

    MODULE = 'Module'
    """A small, typically removable, chassis or card that contains devices
    for a particular subsystem or function."""

    COMPONENT = 'Component'
    """A small chassis, card, or device that contains devices for a
    particular subsystem or function."""

    IP_BASED_DRIVE = 'IPBasedDrive'
    """A chassis in a drive form factor with IP-based network connections."""

    RACK_GROUP = 'RackGroup'
    """A group of racks that form a single entity or share infrastructure."""

    STORAGE_ENCLOSURE = 'StorageEnclosure'
    """A chassis that encloses storage."""

    OTHER = 'Other'
    """A chassis that does not fit any of these definitions."""


# Backward compatibility
CHASSIS_TYPE_RACK = ChassisType.RACK
CHASSIS_TYPE_BLADE = ChassisType.BLADE
CHASSIS_TYPE_ENCLOSURE = ChassisType.ENCLOSURE
CHASSIS_TYPE_STAND_ALONE = ChassisType.STAND_ALONE
CHASSIS_TYPE_RACK_MOUNT = ChassisType.RACK_MOUNT
CHASSIS_TYPE_CARD = ChassisType.CARD
CHASSIS_TYPE_CARTRIDGE = ChassisType.CARTRIDGE
CHASSIS_TYPE_ROW = ChassisType.ROW
CHASSIS_TYPE_POD = ChassisType.POD
CHASSIS_TYPE_EXPANSION = ChassisType.EXPANSION
CHASSIS_TYPE_SIDECAR = ChassisType.SIDECAR
CHASSIS_TYPE_ZONE = ChassisType.ZONE
CHASSIS_TYPE_SLED = ChassisType.SLED
CHASSIS_TYPE_SHELF = ChassisType.SHELF
CHASSIS_TYPE_DRAWER = ChassisType.DRAWER
CHASSIS_TYPE_MODULE = ChassisType.MODULE
CHASSIS_TYPE_COMPONENT = ChassisType.COMPONENT
CHASSIS_TYPE_IP_BASED_DRIVE = ChassisType.IP_BASED_DRIVE
CHASSIS_TYPE_RACK_GROUP = ChassisType.RACK_GROUP
CHASSIS_TYPE_STORAGE_ENCLOSURE = ChassisType.STORAGE_ENCLOSURE
CHASSIS_TYPE_OTHER = ChassisType.OTHER


class IntrusionSensor(enum.Enum):
    """Chassis IntrusionSensor constants"""

    NORMAL = 'Normal'
    """No abnormal physical security condition is detected at this time."""

    HARDWARE_INTRUSION = 'HardwareIntrusion'
    """A door, lock, or other mechanism protecting the internal system
    hardware from being accessed is detected to be in an insecure state."""

    TAMPERING_DETECTED = 'TamperingDetected'
    """Physical tampering of the monitored entity is detected."""


# Backward compatibility
CHASSIS_INTRUSION_SENSOR_NORMAL = IntrusionSensor.NORMAL
CHASSIS_INTRUSION_SENSOR_HARDWARE_INTRUSION = \
    IntrusionSensor.HARDWARE_INTRUSION
CHASSIS_INTRUSION_SENSOR_TAMPERING_DETECTED = \
    IntrusionSensor.TAMPERING_DETECTED


class IntrusionSensorReArm(enum.Enum):
    """Chassis IntrusionSensorReArm constants"""

    MANUAL = 'Manual'
    """A manual re-arm of this sensor restores it to the normal state."""

    AUTOMATIC = 'Automatic'
    """Because no abnormal physical security condition is detected, this
    sensor is automatically restored to the normal state."""


# Backward compatibility
CHASSIS_INTRUSION_SENSOR_RE_ARM_MANUAL = IntrusionSensorReArm.MANUAL
CHASSIS_INTRUSION_SENSOR_RE_ARM_AUTOMATIC = IntrusionSensorReArm.AUTOMATIC



#    Copyright (c) 2021 Anexia Internetdienstleistungs GmbH
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values comes from the Redfish System json-schema:
# https://redfish.dmtf.org/schemas/v1/NetworkDeviceFunction.v1_7_0.json
# https://redfish.dmtf.org/schemas/v1/NetworkPort.v1_4_1.json
# Using Network prefix since some names are too generic.

import enum


class NetworkAuthenticationMethod(enum.Enum):

    NONE = 'None'
    """No iSCSI authentication is used."""

    CHAP = 'CHAP'
    """iSCSI Challenge Handshake Authentication Protocol (CHAP)
    authentication is used."""

    MUTUAL_CHAP = 'MutualCHAP'
    """iSCSI Mutual Challenge Handshake Authentication Protocol (CHAP)
    authentication is used."""


class NetworkBootMode(enum.Enum):

    DISABLED = 'Disabled'
    """Do not indicate to UEFI/BIOS that this device is bootable."""

    PXE = 'PXE'
    """Boot this device by using the embedded PXE support.  Only applicable
    if the NetDevFuncType is `Ethernet` or `InfiniBand`."""

    SCSI = 'iSCSI'
    """Boot this device by using the embedded iSCSI boot support and
    configuration.  Only applicable if the NetDevFuncType is `iSCSI` or
    `Ethernet`."""

    FIBRE_CHANNEL = 'FibreChannel'
    """Boot this device by using the embedded Fibre Channel support and
    configuration.  Only applicable if the NetDevFuncType is
    `FibreChannel`."""

    FIBRE_CHANNEL_OVER_ETHERNET = 'FibreChannelOverEthernet'
    """Boot this device by using the embedded Fibre Channel over Ethernet
    (FCoE) boot support and configuration.  Only applicable if the
    NetDevFuncType is `FibreChannelOverEthernet`."""


class IPAddressType(enum.Enum):

    IPV4 = 'IPv4'
    """IPv4 addressing is used for all IP-fields in this object."""

    IPV6 = 'IPv6'
    """IPv6 addressing is used for all IP-fields in this object."""


class FlowControl(enum.Enum):

    NONE = 'None'
    """No IEEE 802.3x flow control is enabled on this port."""

    TX = 'TX'
    """This station can initiate IEEE 802.3x flow control."""

    RX = 'RX'
    """The link partner can initiate IEEE 802.3x flow control."""

    TX_RX = 'TX_RX'
    """This station or the link partner can initiate IEEE 802.3x flow
    control."""


class NetworkDeviceTechnology(enum.Enum):

    DISABLED = 'Disabled'
    """Neither enumerated nor visible to the operating system."""

    ETHERNET = 'Ethernet'
    """Appears to the operating system as an Ethernet device."""

    FIBRE_CHANNEL = 'FibreChannel'
    """Appears to the operating system as a Fibre Channel device."""

    iSCSI = 'iSCSI'
    """Appears to the operating system as an iSCSI device."""

    FIBRE_CHANNEL_OVER_ETHERNET = 'FibreChannelOverEthernet'
    """Appears to the operating system as an FCoE device."""

    INFINI_BAND = 'InfiniBand'
    """Appears to the operating system as an InfiniBand device."""


class LinkStatus(enum.Enum):

    DOWN = 'Down'
    """The port is enabled but link is down."""

    UP = 'Up'
    """The port is enabled and link is good (up)."""

    STARTING = 'Starting'
    """This link on this interface is starting.  A physical link has been
    established, but the port is not able to transfer data."""

    TRAINING = 'Training'
    """This physical link on this interface is training."""


class PortLinkStatus(enum.Enum):

    LINKUP = 'LinkUp'
    """The port is enabled and link is good (up)."""

    LINKDOWN = 'LinkDown'
    """The port is enabled but link is down."""

    NOLINK = 'NoLink'
    """The port is enabled but have no connectivity."""

    STARTING = 'Starting'
    """This link on this interface is starting.  A physical link has been
    established, but the port is not able to transfer data."""

    TRAINING = 'Training'
    """This physical link on this interface is training."""


class IEEE802IdSubtype(enum.Enum):
    """IEEE 802.1AB ID Subtypes for Chassis ID and Port ID.

    Based on IEEE 802.1AB-2009 and DMTF Redfish Port Schema v1.12.0
    """

    CHASSIS_COMP = 'ChassisComp'
    """Chassis component, based on entPhysicalAlias in RFC4133."""

    IF_ALIAS = 'IfAlias'
    """Interface alias, based on the ifAlias MIB object."""

    PORT_COMP = 'PortComp'
    """Port component, based on entPhysicalAlias in RFC4133."""

    MAC_ADDR = 'MacAddr'
    """MAC address, based on agent-detected unicast source (IEEE 802)."""

    NETWORK_ADDR = 'NetworkAddr'
    """Network address, based on agent-detected network address."""

    IF_NAME = 'IfName'
    """Interface name, based on the ifName MIB object."""

    AGENT_ID = 'AgentId'
    """Agent circuit ID, based on agent-local identifier (RFC3046)."""

    LOCAL_ASSIGN = 'LocalAssign'
    """Locally assigned, based on alphanumeric value."""

    NOT_TRANSMITTED = 'NotTransmitted'
    """No data to be sent to/received from remote partner."""


class LLDPSystemCapabilities(enum.Enum):
    """LLDP System Capabilities.

    Based on IEEE 802.1AB and DMTF Redfish Port Schema v1.12.0
    """

    NONE = 'None'
    """System capabilities are transmitted, but no capabilities are set."""

    BRIDGE = 'Bridge'
    """'bridge' capability."""

    DOCSIS_CABLE_DEVICE = 'DOCSISCableDevice'
    """'DOCSIS cable device' capability."""

    OTHER = 'Other'
    """'other' capability."""

    REPEATER = 'Repeater'
    """'repeater' capability."""

    ROUTER = 'Router'
    """'router' capability."""

    STATION = 'Station'
    """'station' capability."""

    TELEPHONE = 'Telephone'
    """'telephone' capability."""

    WLAN_ACCESS_POINT = 'WLANAccessPoint'
    """'WLAN access point' capability."""



# Copyright (c) 2020-2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import secrets

from sushy import exceptions
from sushy.oem.dell.resources.manager import constants as mgr_cons
from sushy.resources import base
from sushy.resources import common

LOG = logging.getLogger(__name__)


class ForceActionField(base.CompositeField):
    target_uri = base.Field('target', required=True)
    allowed_values = base.Field('Force@Redfish.AllowableValues',
                                adapter=list)


class ActionsField(base.CompositeField):
    reset_idrac = ForceActionField('#DelliDRACCardService.iDRACReset')
    get_kvm_session = common.ActionField('#DelliDRACCardService.GetKVMSession')


class DelliDRACCardService(base.ResourceBase):

    _actions = ActionsField('Actions')
    identity = base.Field('Id', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a DelliDRACCardService.

        :param connector: A Connector instance
        :param identity: The identity of the DelliDRACCardService resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages.
        """
        super().__init__(
            connector, identity, redfish_version, registries)

    def get_allowed_reset_idrac_values(self):
        """Get the allowed values for resetting the idrac.

        :returns: A set of allowed values.
        """
        reset_action = self._actions.reset_idrac

        if not reset_action.allowed_values:
            LOG.warning('Could not figure out the allowed values for the '
                        'reset idrac action for %s', self.identity)
            return set(mgr_cons.ResetType)

        return {v for v in mgr_cons.ResetType
                if v.value in reset_action.allowed_values}

    def reset_idrac(self):
        """Reset the iDRAC.

        """
        reset_type = mgr_cons.ResetType.GRACEFUL
        valid_resets = self.get_allowed_reset_idrac_values()
        if reset_type not in valid_resets:
            raise exceptions.InvalidParameterValueError(
                parameter='value', value=reset_type, valid_values=valid_resets)
        target_uri = self._actions.reset_idrac.target_uri
        payload = {"Force": reset_type.value}
        LOG.debug('Resetting the iDRAC %s ...', self.identity)
        self._conn.post(target_uri, data=payload)
        LOG.info('The iDRAC %s is being reset', self.identity)

    def get_kvm_session(self):
        """Get temporary credentials for KVM session

        The TempUsername and TempPassword fields can be used in the following
        url template:
        https://{host}/console?username={}&tempUsername={}&tempPassword={}
        The username is the user used to generate these session-credentials.

        :returns: Dict with the fields TempUsername and TempPassword as strings
                  None, if the API did not return any credentials, but did not
                  raise an error. When and why that should happen is unclear,
                  but specified in the API doc.
        """
        target_uri = self._actions.get_kvm_session.target_uri
        LOG.debug('Getting KVM session from iDRAC %s ...', self.identity)

        # SessionTypeName: A random string value upto 32 bytes.
        name = secrets.token_urlsafe(32)
        data = {"SessionTypeName": name}
        result = self._conn.post(target_uri, data=data)
        if result.status_code in (200, 201):
            return result.json()
        return None



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from http import client as http_client
import json
import types
from unittest import mock

import sushy
from sushy import exceptions
from sushy.resources.certificateservice import certificate
from sushy.resources.manager import constants as mgr_cons
from sushy.resources.manager import virtual_media
from sushy.tests.unit import base


class _FakeHTTPError(Exception):
    """Mimics the object passed to is_credentials_required()."""

    def __init__(self, status=400, json_body=None, detail=''):
        super().__init__(detail)
        self.detail = detail
        self.response = types.SimpleNamespace(
            status_code=status,
            json=lambda: json_body if json_body is not None else {}
        )


class VirtualMediaTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD'}
        with open('sushy/tests/unit/json_samples/'
                  'virtual_media.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.sys_virtual_media = virtual_media.VirtualMedia(
            self.conn, '/redfish/v1/Managers/BMC/VirtualMedia/Floppy1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sys_virtual_media._parse_attributes(self.json_doc)
        self.assertEqual('Virtual Removable Media',
                         self.sys_virtual_media.name)
        self.assertEqual('Floppy1', self.sys_virtual_media.identity)
        self.assertEqual('https://www.dmtf.org/freeImages/Sardine.img',
                         self.sys_virtual_media.image)
        self.assertEqual('Sardine2.1.43.35.6a',
                         self.sys_virtual_media.image_name)
        self.assertEqual(sushy.ConnectedVia.URI,
                         self.sys_virtual_media.connected_via)
        self.assertEqual([sushy.VirtualMediaType.FLOPPY,
                          sushy.VirtualMediaType.USB_STICK],
                         self.sys_virtual_media.media_types)
        self.assertEqual(True, self.sys_virtual_media.inserted)
        self.assertEqual(False, self.sys_virtual_media.write_protected)

    def test__parse_attributes_return(self):
        attributes = self.sys_virtual_media._parse_attributes(self.json_doc)

        # Test that various types are returned correctly
        self.assertEqual('https://www.dmtf.org/freeImages/Sardine.img',
                         attributes.get('image'))
        self.assertEqual(sushy.ConnectedVia.URI,
                         attributes.get('connected_via'))
        self.assertEqual([sushy.VirtualMediaType.FLOPPY,
                          sushy.VirtualMediaType.USB_STICK],
                         attributes.get('media_types'))

    def test_insert_media_none(self):
        self.sys_virtual_media._actions.insert_media = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #VirtualMedia.InsertMedia',
            self.sys_virtual_media.insert_media,
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)

        self.sys_virtual_media._actions = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #VirtualMedia.InsertMedia',
            self.sys_virtual_media.insert_media,
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)

    def test_insert_media(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)
        self.sys_virtual_media._conn.post.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
             "/VirtualMedia.InsertMedia"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "WriteProtected": False}
        )
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_rf_default(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, True)
        self.sys_virtual_media._conn.post.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
             "/VirtualMedia.InsertMedia"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img"}
        )
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_credentials(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False,
            username="admin", password="pwd")
        self.sys_virtual_media._conn.post.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
             "/VirtualMedia.InsertMedia"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "WriteProtected": False,
                  "UserName": "admin", "Password": "pwd"}
        )
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_credentials_required(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        with open('sushy/tests/unit/json_samples/'
                  'credentials_required_error.json') as f:
            response_obj = json.load(f)
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('GET', 'VirtualMedia', response)

        self.sys_virtual_media._conn.post.side_effect = [error, None]

        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)

        self.sys_virtual_media._conn.post.assert_called_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
             "/VirtualMedia.InsertMedia"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "WriteProtected": False, "UserName": "none",
                  "Password": "none"}
        )
        self.assertEqual(self.sys_virtual_media._conn.post.call_count, 2)
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_transfer_method(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False,
            transfer_method=mgr_cons.TransferMethod.UPLOAD)
        self.sys_virtual_media._conn.post.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
             "/VirtualMedia.InsertMedia"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "WriteProtected": False, "TransferMethod": "Upload"}
        )
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_transfer_method_wrong_input(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.sys_virtual_media.insert_media,
                          "https://www.dmtf.org/freeImages/Sardine.img",
                          True, False, transfer_method='quickly!')

    def test_insert_media_fallback(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH'}
        self.sys_virtual_media._actions.insert_media = None
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "Inserted": True, "WriteProtected": False},
            headers=None)
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_fallback_with_etag(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH',
                                              'ETag': '"3d7b8a7360bf2941d"'}
        self.sys_virtual_media._actions.insert_media = None
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "Inserted": True, "WriteProtected": False},
            headers={"If-Match": '"3d7b8a7360bf2941d"'})
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_insert_media_fallback_with_weak_etag(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH',
                                              'ETag': 'W/"3d7b8a7360bf2941d"'}
        self.sys_virtual_media._actions.insert_media = None
        self.sys_virtual_media.insert_media(
            "https://www.dmtf.org/freeImages/Sardine.img", True, False)
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": "https://www.dmtf.org/freeImages/Sardine.img",
                  "Inserted": True, "WriteProtected": False},
            headers={"If-Match": 'W/"3d7b8a7360bf2941d"'})
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_is_transfer_protocol_required(self):
        with open('sushy/tests/unit/json_samples/'
                  'transfer_proto_required_error.json') as f:
            response_obj = json.load(f)
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('GET', 'VirtualMedia', response)
        retval = self.sys_virtual_media.is_transfer_protocol_required(error)
        self.assertTrue(retval)

    def test_is_transfer_protocol_required_alt(self):
        with open('sushy/tests/unit/json_samples/'
                  'transfer_proto_required_error.json') as f:
            response_obj = json.load(f)
        response_obj['error']['code'] = 'Base.1.5.PropertyMissing'
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('GET', 'VirtualMedia', response)
        retval = self.sys_virtual_media.is_transfer_protocol_required(error)
        self.assertTrue(retval)

    def test_is_transfer_protocol_required_alt2(self):
        with open('sushy/tests/unit/json_samples/'
                  'transfer_proto_required_error2.json') as f:
            response_obj = json.load(f)
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('GET', 'VirtualMedia', response)
        retval = self.sys_virtual_media.is_transfer_protocol_required(error)
        self.assertTrue(retval)

    def test_is_transfer_protocol_required_alt3(self):
        with open('sushy/tests/unit/json_samples/'
                  'transfer_proto_required_error3.json') as f:
            response_obj = json.load(f)
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('GET', 'VirtualMedia', response)
        retval = self.sys_virtual_media.is_transfer_protocol_required(error)
        self.assertTrue(retval)

    def test_is_transfer_method_required(self):
        with open('sushy/tests/unit/json_samples/'
                  'transfer_method_required_error.json') as f:
            response_obj = json.load(f)
        response = mock.Mock(spec=['json', 'status_code'])
        response.json.return_value = response_obj
        error = exceptions.HTTPError('POST', 'VirtualMedia', response)
        retval = self.sys_virtual_media.is_transfer_method_required(error)
        self.assertTrue(retval)

    def test_eject_media_none(self):
        self.sys_virtual_media._actions.eject_media = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #VirtualMedia.EjectMedia',
            self.sys_virtual_media.eject_media)

        self.sys_virtual_media._actions = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #VirtualMedia.EjectMedia',
            self.sys_virtual_media.eject_media)

    def test_eject_media(self):
        self.assertFalse(self.sys_virtual_media._is_stale)
        self.sys_virtual_media.eject_media()
        self.sys_virtual_media._conn.post.assert_called_once_with(
            "/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
            "/VirtualMedia.EjectMedia")
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_fallback(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH'}
        self.sys_virtual_media._actions.eject_media = None
        self.sys_virtual_media.eject_media()
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": None, "Inserted": False}, headers=None)
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_fallback_with_etag(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH',
                                              'ETag': '"3d7b8a7360bf2941d"'}
        self.sys_virtual_media._actions.eject_media = None
        self.sys_virtual_media.eject_media()
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": None, "Inserted": False},
            headers={"If-Match": '"3d7b8a7360bf2941d"'})
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_fallback_with_weak_etag(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD,PATCH',
                                              'ETag': 'W/"3d7b8a7360bf2941d"'}
        self.sys_virtual_media._actions.eject_media = None
        self.sys_virtual_media.eject_media()
        self.sys_virtual_media._conn.patch.assert_called_once_with(
            ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1"),
            data={"Image": None, "Inserted": False},
            headers={"If-Match": 'W/"3d7b8a7360bf2941d"'})
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_pass_empty_dict_415(self):
        target_uri = ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
                      "/VirtualMedia.EjectMedia")
        self.conn.post.side_effect = [exceptions.HTTPError(
            method='POST', url=target_uri, response=mock.MagicMock(
                status_code=http_client.UNSUPPORTED_MEDIA_TYPE)), '200']
        self.sys_virtual_media.eject_media()
        post_calls = [
            mock.call(target_uri),
            mock.call(target_uri, data={})]
        self.sys_virtual_media._conn.post.assert_has_calls(post_calls)
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_pass_empty_dict_400(self):
        target_uri = ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
                      "/VirtualMedia.EjectMedia")
        self.conn.post.side_effect = [exceptions.HTTPError(
            method='POST', url=target_uri, response=mock.MagicMock(
                status_code=http_client.BAD_REQUEST)), '200']
        self.sys_virtual_media.eject_media()
        post_calls = [
            mock.call(target_uri),
            mock.call(target_uri, data={})]
        self.sys_virtual_media._conn.post.assert_has_calls(post_calls)
        self.assertTrue(self.sys_virtual_media._is_stale)

    def test_eject_media_fail_but_media_is_ejected(self):
        target_uri = ("/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Actions"
                      "/VirtualMedia.EjectMedia")
        self.conn.post.side_effect = [
            exceptions.HTTPError(
                method='POST', url=target_uri, response=mock.MagicMock(
                    status_code=http_client.UNSUPPORTED_MEDIA_TYPE)),
            exceptions.HTTPError(
                method='POST', url=target_uri, response=mock.MagicMock(
                    status_code=http_client.INTERNAL_SERVER_ERROR))
        ]

        self.sys_virtual_media.inserted = True

        def fake_refresh(force=False):
            self.sys_virtual_media.inserted = False

        self.sys_virtual_media.refresh = mock.Mock(side_effect=fake_refresh)

        self.sys_virtual_media.eject_media()
        post_calls = [
            mock.call(target_uri),
            mock.call(target_uri, data={})]

        self.sys_virtual_media.refresh.assert_called_once_with(force=True)
        self.sys_virtual_media._conn.post.assert_has_calls(post_calls)

    def test_set_verify_certificate(self):
        self.conn.get.return_value.headers = {'Allow': 'GET,HEAD',
                                              'ETag': '3d7b8a7360bf2941d'}
        with mock.patch.object(
                self.sys_virtual_media, 'invalidate',
                autospec=True) as invalidate_mock:
            self.sys_virtual_media.set_verify_certificate(True)
            self.sys_virtual_media._conn.patch.assert_called_once_with(
                "/redfish/v1/Managers/BMC/VirtualMedia/Floppy1",
                data={'VerifyCertificate': True},
                etag='3d7b8a7360bf2941d')

            invalidate_mock.assert_called_once_with()

    def test_set_verify_certificate_wrong_input(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.sys_virtual_media.set_verify_certificate, 'yes')

    @mock.patch.object(certificate, 'CertificateCollection', autospec=True)
    def test_certificate_collection(self, mock_cert_coll):
        result = self.sys_virtual_media.certificates
        self.assertEqual(mock_cert_coll.return_value, result)
        mock_cert_coll.assert_called_once_with(
            self.conn,
            "/redfish/v1/Managers/BMC/VirtualMedia/Floppy1/Certificates/",
            redfish_version='1.0.2',
            registries=self.sys_virtual_media.registries,
            root=self.sys_virtual_media.root)

    def test_detects_extended_info_username(self):
        body = {
            "error": {
                "@Message.ExtendedInfo": [{
                    "MessageId": "Base.1.12.ActionParameterMissing",
                    "Message": "The action requires the parameter UserName to "
                               "be present."
                }]
            }
        }
        err = _FakeHTTPError(json_body=body)
        self.assertTrue(self.sys_virtual_media.is_credentials_required(err))

    def test_detects_extended_info_password(self):
        body = {
            "error": {
                "@Message.ExtendedInfo": [{
                    "MessageId": "Base.1.15.ActionParameterMissing",
                    "Message": "The action requires the parameter Password to "
                               "be present."
                }]
            }
        }
        err = _FakeHTTPError(json_body=body)
        self.assertTrue(self.sys_virtual_media.is_credentials_required(err))

    def test_detects_plain_message_when_code_absent(self):
        body = {"error": {
            "message": "InsertMedia requires the parameter UserName"}}
        err = _FakeHTTPError(json_body=body, detail="")
        self.assertTrue(self.sys_virtual_media.is_credentials_required(err))

    def test_detects_detail_fallback(self):
        err = _FakeHTTPError(json_body={},
                             detail="requires the parameter Password")
        self.assertTrue(self.sys_virtual_media.is_credentials_required(err))

    def test_preserves_legacy_code_based_detection(self):
        body = {
            "error": {
                "code": "Base.1.10.GeneralError",
                "message": "Action parameter missing: UserName"
            }
        }
        err = _FakeHTTPError(json_body=body)
        self.assertTrue(self.sys_virtual_media.is_credentials_required(err))

    def test_no_trigger_on_unrelated_parameter(self):
        body = {
            "error": {
                "@Message.ExtendedInfo": [{
                    "MessageId": "Base.1.12.ActionParameterMissing",
                    "Message": "The action requires the parameter Timeout to "
                               "be present."
                }]
            }
        }
        err = _FakeHTTPError(json_body=body)
        self.assertFalse(self.sys_virtual_media.is_credentials_required(err))

    def test_no_trigger_on_unrelated_error(self):
        body = {
            "error": {
                "@Message.ExtendedInfo": [{
                    "MessageId": "Base.1.12.ResourceMissingAtURI",
                    "Message": "Resource missing"
                }]
            }
        }
        err = _FakeHTTPError(json_body=body)
        self.assertFalse(self.sys_virtual_media.is_credentials_required(err))

    def test_no_trigger_when_no_clues(self):
        err = _FakeHTTPError(json_body={"error": {"message": "Bad Request"}},
                             detail="")
        self.assertFalse(self.sys_virtual_media.is_credentials_required(err))



# Copyright (c) 2021 Dell Inc. or its subsidiaries.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock

from oslotest.base import BaseTestCase

from sushy.oem.dell.resources.taskservice import constants as ts_cons
from sushy.oem.dell.resources.taskservice import task
from sushy.resources.taskservice import task as sushy_task


class TaskTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/oem/dell/unit/json_samples/'
                  'task.json') as f:
            mock_response = self.conn.get.return_value
            mock_response.json.return_value = json.load(f)
            mock_response.status_code = 200

        self.task = sushy_task.Task(
            self.conn,
            '/redfish/v1/TaskService/Tasks/JID_257309938313')
        self.oem_task = task.DellTaskExtension(
            self.conn,
            '/redfish/v1/TaskService/Tasks/JID_257309938313')
        self.oem_task = self.oem_task.set_parent_resource(
            self.task, 'Dell')

    def test_parse_attributes(self):
        self.assertEqual('JID_257309938313', self.oem_task.identity)
        self.assertEqual('Configure: RAID.Integrated.1-1',
                         self.oem_task.name)
        self.assertEqual('Job Instance', self.oem_task.description)
        self.assertIsNone(self.oem_task.completion_time)
        self.assertEqual('TIME_NA', self.oem_task.end_time)
        self.assertEqual(ts_cons.JobState.SCHEDULED,
                         self.oem_task.job_state)
        self.assertEqual(ts_cons.JobType.RAID_CONF,
                         self.oem_task.job_type)
        # For backward compatibility
        self.assertEqual(ts_cons.JOB_STATE_SCHEDULED,
                         self.oem_task.job_state)
        self.assertEqual(ts_cons.JOB_TYPE_RAID_CONF,
                         self.oem_task.job_type)
        self.assertEqual('Task successfully scheduled.',
                         self.oem_task.message)
        self.assertEqual([], self.oem_task.message_args)
        self.assertEqual('IDRAC.2.5.JCP001', self.oem_task.message_id)
        self.assertEqual(0, self.oem_task.percent_complete)
        self.assertEqual('TIME_NOW', self.oem_task.start_time)



#    Copyright (c) 2021 Anexia Internetdienstleistungs GmbH
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/NetworkPort.v1_2_1.json

from sushy.resources import base
from sushy.resources import common
from sushy.resources.system.network import constants


class NetworkPortCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return NetworkPort


class NetworkPort(base.ResourceBase):
    associated_network_addresses = base.Field(
        "AssociatedNetworkAddresses", adapter=list
    )
    """The array of configured network addresses that are associated."""

    current_link_speed_mbps = base.Field(
        "CurrentLinkSpeedMbps", adapter=int
    )
    """The network port current link speed."""

    description = base.Field('Description')
    """The network port description"""

    flow_control_configuration = base.MappedField('FlowControlConfiguration',
                                                  constants.FlowControl)
    """The locally configured 802.3x flow control setting."""

    flow_control_status = base.MappedField('FlowControlStatus',
                                           constants.FlowControl)
    """The 802.3x flow control behavior negotiated with the link partner"""

    identity = base.Field('Id', required=True)
    """The network port identity"""

    link_status = base.MappedField('LinkStatus', constants.LinkStatus)
    """The link status of the network port."""

    name = base.Field(
        "Name", required=True
    )
    """The network port name"""

    physical_port_number = base.Field("PhysicalPortNumber", adapter=int)
    """The physical port number label for this port."""

    status = common.StatusField('Status')
    """The network port status"""



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


from sushy.resources import constants as res_cons
from sushy.resources.system import simple_storage
from sushy.tests.unit import base


class SimpleStorageTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.simpl_stor = simple_storage.SimpleStorage(
            self.conn, '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.simpl_stor._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.simpl_stor.redfish_version)
        self.assertEqual('1', self.simpl_stor.identity)
        self.assertEqual('Simple Storage Controller', self.simpl_stor.name)
        self.assertEqual(8000000000000,
                         self.simpl_stor.devices[0].capacity_bytes)
        self.assertEqual(4000000000000,
                         self.simpl_stor.devices[1].capacity_bytes)
        self.assertEqual(res_cons.State.ENABLED,
                         self.simpl_stor.devices[0].status.state)
        self.assertEqual(res_cons.State.ABSENT,
                         self.simpl_stor.devices[2].status.state)
        self.assertEqual(res_cons.Health.OK,
                         self.simpl_stor.devices[0].status.health)


class SimpleStorageCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage_collection.json') as f:

            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.simpl_stor_col = simple_storage.SimpleStorageCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/SimpleStorage',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.simpl_stor_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',),
            self.simpl_stor_col.members_identities)

    @mock.patch.object(simple_storage, 'SimpleStorage', autospec=True)
    def test_get_member(self, SimpleStorage_mock):
        self.simpl_stor_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1')
        SimpleStorage_mock.assert_called_once_with(
            self.simpl_stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            redfish_version=self.simpl_stor_col.redfish_version,
            registries=None,
            root=self.simpl_stor_col.root)

    @mock.patch.object(simple_storage, 'SimpleStorage', autospec=True)
    def test_get_members(self, SimpleStorage_mock):
        members = self.simpl_stor_col.get_members()
        SimpleStorage_mock.assert_called_once_with(
            self.simpl_stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            redfish_version=self.simpl_stor_col.redfish_version,
            registries=None, root=self.simpl_stor_col.root)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))

    def test_disks_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual([4000000000000, 8000000000000],
                         self.simpl_stor_col.disks_sizes_bytes)

    def test_disks_sizes_bytes_capacity_bytes_none(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            json_doc = json.load(f)

        json_doc['Devices'][0]['CapacityBytes'] = None
        self.conn.get.return_value.json.return_value = json_doc

        self.assertEqual([4000000000000],
                         self.simpl_stor_col.disks_sizes_bytes)

    def test_max_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_size_bytes_after_refresh(self):
        self.simpl_stor_col.refresh()
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)

    def test_get_members_expanded(self):
        # Test with expanded JSON data containing full member objects
        # This simulates a response from ?$expand=.($levels=1)
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage_collection_expanded.json') as f:
            expanded_json = json.loads(f.read())

        # Create collection with the URL that would have ?$expand appended
        expanded_collection = simple_storage.SimpleStorageCollection(
            self.conn,
            "/redfish/v1/Systems/1/SimpleStorage?$expand=.($levels=1)",
            redfish_version="1.0.2",
        )
        expanded_collection._json = expanded_json

        # Mock the SimpleStorage objects that will be created
        mock_member1 = mock.Mock(spec=simple_storage.SimpleStorage)
        mock_member1.identity = "RAID-1"
        mock_member1.name = "RAID Controller"

        mock_member2 = mock.Mock(spec=simple_storage.SimpleStorage)
        mock_member2.identity = "AHCI-1"
        mock_member2.name = "AHCI Controller"

        with mock.patch.object(
            simple_storage, "SimpleStorage", autospec=True
        ) as mock_simple_storage:
            # Make the constructor return our mocked instances
            mock_simple_storage.side_effect = [mock_member1, mock_member2]

            members = expanded_collection.get_members()

            # Verify it created SimpleStorage objects from expanded data
            self.assertEqual(2, len(members))
            self.assertEqual([mock_member1, mock_member2], members)

            # Check that SimpleStorage was initialized with expanded JSON data
            self.assertEqual(2, mock_simple_storage.call_count)

            # Verify first member initialization with expanded data
            first_call = mock_simple_storage.call_args_list[0]
            self.assertEqual(expanded_collection._conn, first_call[0][0])
            self.assertEqual(
                "/redfish/v1/Systems/1/SimpleStorage/RAID-1",
                first_call[0][1],
            )
            # Check that expanded data was passed
            self.assertEqual(
                expanded_json["Members"][0], first_call[1]["json_doc"]
            )

            # Verify second member initialization with expanded data
            second_call = mock_simple_storage.call_args_list[1]
            self.assertEqual(expanded_collection._conn, second_call[0][0])
            self.assertEqual(
                "/redfish/v1/Systems/1/SimpleStorage/AHCI-1",
                second_call[0][1],
            )
            # Check that expanded data was passed
            self.assertEqual(
                expanded_json["Members"][1], second_call[1]["json_doc"]
            )

    def test_get_members_unexpanded(self):
        # Test with unexpanded JSON data containing only references
        # This simulates a normal response without ?$expand parameter
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage_collection.json') as f:
            unexpanded_json = json.loads(f.read())

        # Create collection without expand in URL
        unexpanded_collection = simple_storage.SimpleStorageCollection(
            self.conn,
            "/redfish/v1/Systems/1/SimpleStorage",
            redfish_version="1.0.2",
        )
        unexpanded_collection._json = unexpanded_json

        # Mock the parent get_members method which will fetch each member
        # individually
        mock_member1 = mock.Mock(spec=simple_storage.SimpleStorage)
        mock_member2 = mock.Mock(spec=simple_storage.SimpleStorage)
        with mock.patch.object(
            simple_storage.base.ResourceCollectionBase,
            "get_members",
            autospec=True,
            return_value=[mock_member1, mock_member2],
        ) as mock_super:
            members = unexpanded_collection.get_members()

            # Should fall back to parent implementation for individual fetches
            mock_super.assert_called_once_with(unexpanded_collection)
            self.assertEqual([mock_member1, mock_member2], members)

    def test_get_members_no_json(self):
        # Test when collection has no _json attribute
        collection_no_json = simple_storage.SimpleStorageCollection(
            self.conn,
            "/redfish/v1/Systems/1/SimpleStorage",
            redfish_version="1.0.2",
        )

        # Mock the parent get_members method
        with mock.patch.object(
            simple_storage.base.ResourceCollectionBase,
            "get_members",
            autospec=True,
            return_value=["mock_member"],
        ) as mock_super:
            members = collection_no_json.get_members()

            # Should fall back to parent implementation
            mock_super.assert_called_once()
            self.assertEqual(["mock_member"], members)



# Copyright (c) 2020 Dell Inc. or its subsidiaries.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslotest.base import BaseTestCase

import sushy
from sushy.oem.dell.asynchronous import http_call


class AsychronousTestCase(BaseTestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()

    def test_http_call_post_accepted(self):
        mock_post_response = self.conn.post.return_value
        mock_post_response.status_code = 202
        mock_post_response.headers.get.return_value = '1'

        mock_get_202_response = mock.Mock()
        mock_get_202_response.status_code = 202
        mock_get_202_response.headers.get.return_value = '1'

        mock_get_200_response = mock.Mock()
        mock_get_200_response.status_code = 200

        self.conn.get.side_effect = [
            mock_get_202_response, mock_get_200_response]

        resp = http_call(self.conn, 'POST')
        self.assertIs(resp, mock_get_200_response)

    def test_http_call_post_accepted_no_location(self):
        mock_response = self.conn.post.return_value
        mock_response.status_code = 202
        mock_response.headers.get.return_value = None

        self.assertRaises(sushy.exceptions.ExtensionError,
                          http_call, self.conn, 'POST')

    @mock.patch('time.sleep', autospec=True)
    def test_http_call_404_retry_success(self, mock_sleep):
        # First call returns 404, second call succeeds
        mock_404_response = mock.Mock()
        mock_404_response.status_code = 404

        mock_200_response = mock.Mock()
        mock_200_response.status_code = 200

        self.conn.post.side_effect = [mock_404_response, mock_200_response]

        resp = http_call(self.conn, 'POST', '/some/path')

        self.assertIs(resp, mock_200_response)
        self.assertEqual(self.conn.post.call_count, 2)
        mock_sleep.assert_called_once_with(10)  # Default retry delay

    @mock.patch('time.sleep', autospec=True)
    def test_http_call_404_retry_custom_params(self, mock_sleep):
        # Test with custom retry parameters
        mock_404_response = mock.Mock()
        mock_404_response.status_code = 404

        mock_200_response = mock.Mock()
        mock_200_response.status_code = 200

        self.conn.post.side_effect = [mock_404_response, mock_200_response]

        resp = http_call(self.conn, 'POST', '/some/path',
                         max_404_retries=5, retry_404_delay=2)

        self.assertIs(resp, mock_200_response)
        self.assertEqual(self.conn.post.call_count, 2)
        mock_sleep.assert_called_once_with(2)  # Custom delay

    @mock.patch('time.sleep', autospec=True)
    def test_http_call_404_retry_exhausted(self, mock_sleep):
        # All retries return 404, should fail with ExtensionError
        mock_404_response = mock.Mock()
        mock_404_response.status_code = 404

        self.conn.post.return_value = mock_404_response

        self.assertRaisesRegex(sushy.exceptions.ExtensionError,
                               'failed with code 404',
                               http_call, self.conn, 'POST', '/some/path')

        # 1 initial + 3 retries
        self.assertEqual(self.conn.post.call_count, 4)
        self.assertEqual(mock_sleep.call_count, 3)

    @mock.patch('time.sleep', autospec=True)
    def test_http_call_non_404_error_no_retry(self, mock_sleep):
        # Non-404 errors should not trigger retry
        mock_500_response = mock.Mock()
        mock_500_response.status_code = 500

        self.conn.post.return_value = mock_500_response

        self.assertRaisesRegex(sushy.exceptions.ExtensionError,
                               'failed with code 500',
                               http_call, self.conn, 'POST', '/some/path')

        self.assertEqual(self.conn.post.call_count, 1)  # No retries
        mock_sleep.assert_not_called()

    def test_http_call_success_no_retry(self):
        # Successful response should not trigger any retry logic
        mock_200_response = mock.Mock()
        mock_200_response.status_code = 200

        self.conn.post.return_value = mock_200_response

        resp = http_call(self.conn, 'POST', '/some/path')

        self.assertIs(resp, mock_200_response)
        self.assertEqual(self.conn.post.call_count, 1)

    @mock.patch('time.sleep', autospec=True)
    def test_http_call_404_retry_with_202_polling(self, mock_sleep):
        # Test 404 retry followed by 202 polling behavior
        mock_404_response = mock.Mock()
        mock_404_response.status_code = 404

        mock_202_response = mock.Mock()
        mock_202_response.status_code = 202

        def mock_get_header(header_name, default=None):
            if header_name == 'Location':
                return '/redfish/v1/TaskService/Tasks/123'
            elif header_name == 'Retry-After':
                return None
            return default
        mock_202_response.headers.get.side_effect = mock_get_header

        mock_200_response = mock.Mock()
        mock_200_response.status_code = 200

        self.conn.post.side_effect = [mock_404_response, mock_202_response]
        self.conn.get.return_value = mock_200_response

        resp = http_call(self.conn, 'POST', '/some/path')

        self.assertIs(resp, mock_200_response)
        self.assertEqual(self.conn.post.call_count, 2)
        self.assertEqual(self.conn.get.call_count, 1)
        # Should have slept for 404 retry and 202 polling
        self.assertEqual(mock_sleep.call_count, 2)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from datetime import timedelta
import logging
import time

from dateutil import parser

import sushy

LOG = logging.getLogger(__name__)

TASK_POLL_PERIOD = 1


def _to_datetime(retry_after_str):
    if retry_after_str.isdigit():
        # Retry-After: 120
        return datetime.now() + timedelta(seconds=int(retry_after_str))
    else:
        # Retry-After: Fri, 31 Dec 1999 23:59:59 GMT
        return parser.parse(retry_after_str)


def http_call(conn, method, *args, **kwargs):
    handle = getattr(conn, method.lower())

    sleep_for = kwargs.pop('sushy_task_poll_period', TASK_POLL_PERIOD)
    max_404_retries = kwargs.pop('max_404_retries', 3)
    retry_404_delay = kwargs.pop('retry_404_delay', 10)

    # Retry initial call on 404 to handle BMC race condition
    for attempt in range(max_404_retries + 1):
        response = handle(*args, **kwargs)

        LOG.debug('Finished HTTP %s with args %s %s, response is '
                  '%d', method, args or '', kwargs, response.status_code)

        # If not 404 or last attempt, break out of retry loop
        if response.status_code != 404 or attempt == max_404_retries:
            break

        LOG.warning('BMC returned 404 for HTTP %(method)s request to '
                    '%(path)s - this may be a transient condition while '
                    'BMC creates the job. Retrying in %(delay)d seconds '
                    '(attempt %(current)d of %(total)d)',
                    {'method': method.upper(),
                     'path': args[0] if args else 'unknown',
                     'delay': retry_404_delay,
                     'current': attempt + 1,
                     'total': max_404_retries})
        time.sleep(retry_404_delay)

    location = None
    while response.status_code == 202:
        location = response.headers.get('Location', location)
        if not location:
            raise sushy.exceptions.ExtensionError(
                error='Response %d to HTTP %s with args %s, kwargs %s '
                      'does not include Location: in '
                      'header' % (response.status_code, method.upper(),
                                  args, kwargs))

        retry_after = response.headers.get('Retry-After')
        if retry_after:
            retry_after = _to_datetime(retry_after)
            sleep_for = max(0, (retry_after - datetime.now()).total_seconds())

        LOG.debug('Sleeping for %d secs before retrying HTTP GET '
                  '%s', sleep_for, location)

        time.sleep(sleep_for)

        response = conn.get(location)

        LOG.debug('Finished HTTP GET %s, response is '
                  '%d', location, response.status_code)

    if response.status_code >= 400:
        raise sushy.exceptions.ExtensionError(
            error=f'HTTP {method.upper()} with args {args}, '
                  f'kwargs {kwargs} failed '
                  f'with code {response.status_code}')

    return response



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import collections
from importlib import resources
import logging
import os

import requests

from sushy import auth as sushy_auth
from sushy import connector as sushy_connector
from sushy import exceptions
from sushy.resources import base
from sushy.resources.certificateservice import certificateservice
from sushy.resources.chassis import chassis
from sushy.resources.compositionservice import compositionservice
from sushy.resources.eventservice import eventservice
from sushy.resources.fabric import fabric
from sushy.resources.manager import manager
from sushy.resources.registry import message_registry
from sushy.resources.registry import message_registry_file
from sushy.resources.sessionservice import session
from sushy.resources.sessionservice import sessionservice
from sushy.resources.system import system
from sushy.resources.taskservice import taskservice
from sushy.resources.updateservice import updateservice
from sushy import taskmonitor
from sushy import utils

LOG = logging.getLogger(__name__)

STANDARD_REGISTRY_PATH = 'standard_registries'


class ProtocolFeaturesSupportedField(base.CompositeField):

    excerpt_query = base.Field('ExcerptQuery')
    """The excerpt query parameter is supported"""

    expand_query = base.Field('ExpandQuery')
    """The expand query parameter is supported"""

    filter_query = base.Field('FilterQuery')
    """The filter query parameter is supported"""

    only_member_query = base.Field('OnlyMemberQuery')
    """The only query parameter is supported"""

    select_query = base.Field('SelectQuery')
    """The select query parameter is supported"""


class LazyRegistries(collections.abc.MutableMapping):
    """Download registries on demand.

    Redfish message registries can be very large. On top of that,
    they are not used frequently. Thus, let's not pull them off
    the BMC unless the consumer is actually trying to use them.

    :param service_root: Redfish service root object
    :type service_root: sushy.main.Sushy
    """

    def __init__(self, service_root):
        self._service_root = service_root
        self._registries = None

    def __getitem__(self, key):
        registries = self.registries
        return registries[key]

    def __setitem__(self, key, value):
        registries = self.registries
        registries[key] = value

    def __delitem__(self, key):
        registries = self.registries
        del registries[key]

    def __iter__(self):
        registries = self.registries
        return iter(registries or ())

    def __len__(self):
        registries = self.registries
        return len(registries)

    @property
    def registries(self):
        if self._registries is None:
            self._registries = self._service_root.registries

        return self._registries


class Sushy(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The Redfish root service identity"""

    name = base.Field('Name')
    """The Redfish root service name"""

    uuid = base.Field('UUID')
    """The Redfish root service UUID"""

    product = base.Field('Product')
    """The product associated with this Redfish service"""

    protocol_features_supported = ProtocolFeaturesSupportedField(
        'ProtocolFeaturesSupported')
    """The information about protocol features supported by the service"""

    _composition_service_path = base.Field(
        ['CompositionService', '@odata.id'])
    """CompositionService path"""

    _systems_path = base.Field(['Systems', '@odata.id'])
    """SystemCollection path"""

    _managers_path = base.Field(['Managers', '@odata.id'])
    """ManagerCollection path"""

    _chassis_path = base.Field(['Chassis', '@odata.id'])
    """ChassisCollection path"""

    _fabrics_path = base.Field(['Fabrics', '@odata.id'])
    """FabricCollection path"""

    _session_service_path = base.Field(['SessionService', '@odata.id'])
    """SessionService path"""

    _registries_path = base.Field(['Registries', '@odata.id'])
    """Registries path"""

    _update_service_path = base.Field(['UpdateService', '@odata.id'])
    """UpdateService path"""

    _event_service_path = base.Field(['EventService', '@odata.id'])
    """EventService path"""

    _certificate_service_path = base.Field(['CertificateService', '@odata.id'])
    """CertificateService path"""

    def __init__(self, base_url, username=None, password=None,
                 root_prefix='/redfish/v1/', verify=True,
                 auth=None, connector=None,
                 public_connector=None,
                 language='en', server_side_retries=10,
                 server_side_retries_delay=3,
                 read_timeout=60,
                 connect_timeout=None):
        """A class representing a RootService

        :param base_url: The base URL to the Redfish controller. It
            should include scheme and authority portion of the URL. For
            example: https://mgmt.vendor.com
        :param username: User account with admin/server-profile access
            privilege
        :param password: User account password
        :param root_prefix: The default URL prefix. This part includes
            the root service and version. Defaults to /redfish/v1
        :param verify: Either a boolean value, a path to a CA_BUNDLE
            file or directory with certificates of trusted CAs. If set to
            True the driver will verify the host certificates; if False
            the driver will ignore verifying the SSL certificate; if it's
            a path the driver will use the specified certificate or one of
            the certificates in the directory. Defaults to True.
        :param auth: An authentication mechanism to utilize.
        :param connector: A user-defined connector object. Defaults to None.
        :param public_connector: A user-defined connector to use for requests
            on the Internet, e.g., for Message Registries. Defaults to None.
        :param language: RFC 5646 language code for Message Registries.
            Defaults to 'en'.
        :param server_side_retries: Number of times to retry GET requests in
            case of server side errors. Defaults to 10.
        :param server_side_retries_delay: Time in seconds between retries of
            GET requests in case of server side errors. Defaults to 3.
        :param read_timeout: HTTP read timeout in seconds. This is the maximum
            time to wait for a response from the BMC after the connection is
            established. Defaults to 60.
        :param connect_timeout: TCP connection timeout in seconds. This is
            the maximum time to wait for the initial TCP connection to the
            BMC to be established. If not specified, read_timeout is used for
            both connect and read timeouts. Setting this to a lower value
            (e.g., 10) allows faster failure when a BMC is unreachable.
        """
        self._root_prefix = root_prefix
        if (auth is not None and (password is not None
                                  or username is not None)):
            msg = ('Username or Password were provided to Sushy '
                   'when an authentication mechanism was specified.')
            raise ValueError(msg)
        if auth is None:
            auth = sushy_auth.SessionOrBasicAuth(username=username,
                                                 password=password)
        self._auth = auth

        super().__init__(
            connector or sushy_connector.Connector(
                base_url, verify=verify,
                server_side_retries=server_side_retries,
                server_side_retries_delay=server_side_retries_delay,
                default_request_timeout=read_timeout,
                connect_timeout=connect_timeout),
            path=self._root_prefix)
        self._public_connector = public_connector or requests
        self._language = language
        self._base_url = base_url
        self._auth.set_context(self, self._conn)
        self._auth.authenticate()

    def __del__(self):
        if self._auth:
            try:
                self._auth.close()

            except Exception as ex:
                LOG.warning('Ignoring error while closing Redfish session '
                            'with %s: %s', self._base_url, ex)
            self._auth = None

    def _parse_attributes(self, json_doc):
        """Parse the attributes of a resource.

        Parsed JSON fields are set to `self` as declared in the class.

        :param json_doc: parsed JSON document in form of Python types
        """
        super()._parse_attributes(json_doc)
        self.redfish_version = json_doc.get('RedfishVersion')

    def get_system_collection(self):
        """Get the SystemCollection object

        :raises: MissingAttributeError, if the collection attribute is
            not found
        :returns: a SystemCollection object
        """
        if not self._systems_path:
            raise exceptions.MissingAttributeError(
                attribute='Systems/@odata.id', resource=self._path)

        return system.SystemCollection(
            self._conn, self._systems_path,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def get_system(self, identity=None):
        """Given the identity return a System object

        :param identity: The identity of the System resource. If not given,
            sushy will default to the single available System or fail
            if there appear to be more or less then one System listed.
        :raises: `UnknownDefaultError` if default system can't be determined.
        :returns: The System object
        """
        if identity is None:
            systems_collection = self.get_system_collection()
            listed_systems = systems_collection.get_members()
            if len(listed_systems) != 1:
                raise exceptions.UnknownDefaultError(
                    entity='ComputerSystem',
                    error='System count is not exactly one')

            identity = listed_systems[0].path

        return system.System(self._conn, identity,
                             redfish_version=self.redfish_version,
                             registries=self.lazy_registries, root=self)

    def get_chassis_collection(self):
        """Get the ChassisCollection object

        :raises: MissingAttributeError, if the collection attribute is
            not found
        :returns: a ChassisCollection object
        """
        if not self._chassis_path:
            raise exceptions.MissingAttributeError(
                attribute='Chassis/@odata.id', resource=self._path)

        return chassis.ChassisCollection(self._conn, self._chassis_path,
                                         redfish_version=self.redfish_version,
                                         registries=self.lazy_registries,
                                         root=self)

    def get_chassis(self, identity=None):
        """Given the identity return a Chassis object

        :param identity: The identity of the Chassis resource. If not given,
            sushy will default to the single available chassis or fail
            if there appear to be more or less then one Chassis listed.
        :raises: `UnknownDefaultError` if default system can't be determined.
        :returns: The Chassis object
        """
        if identity is None:
            chassis_collection = self.get_chassis_collection()
            listed_chassis = chassis_collection.get_members()
            if len(listed_chassis) != 1:
                raise exceptions.UnknownDefaultError(
                    entity='Chassis',
                    error='Chassis count is not exactly one')

            identity = listed_chassis[0].path

        return chassis.Chassis(self._conn, identity,
                               redfish_version=self.redfish_version,
                               registries=self.lazy_registries, root=self)

    def get_fabric_collection(self):
        """Get the FabricCollection object

        :raises: MissingAttributeError, if the collection attribute is
            not found
        :returns: a FabricCollection object
        """
        if not self._fabrics_path:
            raise exceptions.MissingAttributeError(
                attribute='Fabrics/@odata.id', resource=self._path)

        return fabric.FabricCollection(self._conn, self._fabrics_path,
                                       redfish_version=self.redfish_version,
                                       registries=self.lazy_registries,
                                       root=self)

    def get_fabric(self, identity):
        """Given the identity return a Fabric object

        :param identity: The identity of the Fabric resource
        :returns: The Fabric object
        """
        return fabric.Fabric(self._conn, identity,
                             redfish_version=self.redfish_version,
                             registries=self.lazy_registries, root=self)

    def get_manager_collection(self):
        """Get the ManagerCollection object

        :raises: MissingAttributeError, if the collection attribute is
            not found
        :returns: a ManagerCollection object
        """
        if not self._managers_path:
            raise exceptions.MissingAttributeError(
                attribute='Managers/@odata.id', resource=self._path)

        return manager.ManagerCollection(self._conn, self._managers_path,
                                         redfish_version=self.redfish_version,
                                         registries=self.lazy_registries,
                                         root=self)

    def get_manager(self, identity=None):
        """Given the identity return a Manager object

        :param identity: The identity of the Manager resource. If not given,
            sushy will default to the single available Manager or fail
            if there appear to be more or less then one Manager listed.
        :returns: The Manager object
        """
        if identity is None:
            managers_collection = self.get_manager_collection()
            listed_managers = managers_collection.get_members()
            if len(listed_managers) != 1:
                raise exceptions.UnknownDefaultError(
                    entity='Manager',
                    error='Manager count is not exactly one')

            identity = listed_managers[0].path

        return manager.Manager(self._conn, identity,
                               redfish_version=self.redfish_version,
                               registries=self.lazy_registries, root=self)

    def get_session_service(self):
        """Get the SessionService object

        :raises: MissingAttributeError, if the collection attribute is
            not found
        :returns: as SessionCollection object
        """
        if not self._session_service_path:
            raise exceptions.MissingAttributeError(
                attribute='SessionService/@odata.id', resource=self._path)

        return sessionservice.SessionService(
            self._conn, self._session_service_path,
            redfish_version=self.redfish_version, root=self)

    def get_sessions_path(self):
        """Returns the Sessions url"""
        # NOTE(TheJulia): This method is the standard discovery method
        # advised by the DMTF DSP0266 standard to find the session service.
        try:
            links_url = self.json.get('Links')
            sessions_uri = links_url['Sessions']['@odata.id']
            # Save the session URL for post detection and prevention
            # of recursive authentication attempts.
            self._conn._sessions_uri = sessions_uri
            return sessions_uri
        except (TypeError, KeyError):
            raise exceptions.MissingAttributeError(
                attribute='Links/Sessions/@data.id', resource=self.path)

    def get_session(self, identity):
        """Given the identity return a Session object

        :param identity: The identity of the session resource
        :returns: The Session object
        """
        return session.Session(
            self._conn, identity,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def create_session(self, username=None, password=None):
        """Creates a session without invoking SessionService.

        For use when a new connection is to be established. Removes prior
        Session and authentication data before making the request.

        :param username: The username to utilize to create a session with the
                         remote endpoint.
        :param password: The password to utilize to create a session with the
                         remote endpoint.
        :returns: A session key and uri in the form of a tuple
        :raises: MissingXAuthToken
        :raises: ConnectionError
        :raises: AccessError
        :raises: HTTPError
        :raises: MissingAttributeError
        """
        # Explicitly removes in-client session data to proceed. This prevents
        # AccessErrors as prior authentication shouldn't be submitted with a
        # new authentication attempt, and doing so with old/invalid session
        # data can result in an AccessError being raised.
        self._conn._session.auth = None
        if 'X-Auth-Token' in self._conn._session.headers:
            # Delete the token value that was saved to the session
            # as otherwise we would end up with a dictionary containing
            # a {'X-Auth-Token': null} being sent across to the remote
            # bmc.
            del self._conn._session.headers['X-Auth-Token']
        try:
            session_service_path = self.get_sessions_path()
        except (exceptions.MissingAttributeError, exceptions.AccessError):
            # Guesses path based upon DMTF standard, in the event
            # the links resource on the service root is incorrect.
            session_service_path = os.path.join(self._path,
                                                'SessionService/Sessions')
            LOG.warning('Could not discover the Session service path, '
                        'falling back to %s.',
                        session_service_path)
            session_url = self._root_prefix + 'SessionService/Sessions'
            self._conn._sessions_uri = session_url

        data = {'UserName': username, 'Password': password}
        LOG.debug("Requesting new session from %s.",
                  session_service_path)
        rsp = self._conn.post(session_service_path, data=data)
        session_key = rsp.headers.get('X-Auth-Token')
        if session_key is None:
            raise exceptions.MissingXAuthToken(
                method='POST', url=session_service_path, response=rsp)

        session_uri = rsp.headers.get('Location')
        if session_uri is None:
            try:
                body = rsp.json()
                session_uri = body.get("@odata.id")
            except ValueError:  # JSON decoding failed
                pass

        if session_uri is None:
            LOG.warning("Received X-Auth-Token but NO session uri.")

        return session_key, session_uri

    def get_update_service(self):
        """Get the UpdateService object

        :returns: The UpdateService object
        """
        if not self._update_service_path:
            raise exceptions.MissingAttributeError(
                attribute='UpdateService/@odata.id', resource=self._path)

        return updateservice.UpdateService(
            self._conn, self._update_service_path,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def get_task_service(self):
        """Get the TaskService object

        :returns: The TaskService object
        """
        return taskservice.TaskService(
            self._conn, utils.get_sub_resource_path_by(self, 'Tasks'),
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def _get_registry_collection(self):
        """Get MessageRegistryFileCollection object

        This resource is optional and can be empty.

        :returns: MessageRegistryFileCollection object
            or None if Registries not provided
        """

        if self._registries_path:
            return message_registry_file.MessageRegistryFileCollection(
                self._conn,
                self._registries_path,
                redfish_version=self.redfish_version, root=self)

    def get_composition_service(self):
        """Get the CompositionService object

        :raises: MissingAttributeError, if the composition service
            attribute is not found
        :returns: The CompositionService object
        """
        if not self._composition_service_path:
            raise exceptions.MissingAttributeError(
                attribute='CompositionService/@odata.id',
                resource=self._path)

        return compositionservice.CompositionService(
            self._conn, self._composition_service_path,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def get_certificate_service(self):
        """Get the CertificateService object

        :returns: The CertificateService object
        """
        if not self._certificate_service_path:
            raise exceptions.MissingAttributeError(
                attribute='CertificateService/@odata.id',
                resource=self._path)

        return certificateservice.CertificateService(
            self._conn, self._certificate_service_path,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def get_event_service(self):
        """Get the EventService object

        :raises: MissingAttributeError, if the EventService is not found
        :returns: The EventService object
        """
        if not self._event_service_path:
            raise exceptions.MissingAttributeError(
                attribute='EventService/@odata.id',
                resource=self._path)
        return eventservice.EventService(
            self._conn, self._event_service_path,
            redfish_version=self.redfish_version,
            registries=self.lazy_registries, root=self)

    def _get_standard_message_registry_collection(self):
        """Load packaged standard message registries

        :returns: list of MessageRegistry
        """

        return [
            message_registry.MessageRegistry(
                None,
                os.path.join(STANDARD_REGISTRY_PATH, json_file.name),
                reader=base.JsonPackagedFileReader(__package__),
            )
            for json_file in resources.files(__package__)
            .joinpath(STANDARD_REGISTRY_PATH)
            .iterdir()
            if json_file.is_file()
        ]

    @property
    @utils.cache_it
    def registries(self):
        """Gets and combines all registries together

        Fetches all registries if any provided by Redfish service
        and combines together with packaged standard registries.
        Both message and attribute registries are supported from
        the Redfish service.

        :returns: dict of combined registries keyed by both the
            registry name (Registry_name.Major_version.Minor_version) and the
            registry file identity, with the value being the actual
            registry itself.
        """
        standard = self._get_standard_message_registry_collection()

        registries = {r.registry_prefix + '.'
                      + r.registry_version.rsplit('.', 1)[0]: r
                      for r in standard if r.language == self._language}

        registry_col = self._get_registry_collection()

        endpoint_registries = {}
        if registry_col:
            provided = registry_col.get_members()
            for r in provided:

                # Check for Message and Attribute registries
                registry = r.get_message_registry(
                    self._language,
                    self._public_connector)
                if not registry:
                    registry = r.get_attribute_registry(
                        self._language,
                        self._public_connector)
                if registry:
                    endpoint_registries[r.registry] = registry
                    endpoint_registries.setdefault(r.identity, registry)

        if endpoint_registries:
            LOG.debug('Found registries for %(id)s: %(reg)s',
                      {'id': self.identity,
                       'reg': ', '.join(endpoint_registries)})
            registries.update(endpoint_registries)
        else:
            LOG.debug('No registries are available for %s', self.identity)

        return registries

    @property
    def lazy_registries(self):
        """Gets and combines all message registries together

        Fetches all registries if any provided by Redfish service
        and combines together with packaged standard registries.

        :returns: dict of combined message registries where key is
            Registry_name.Major_version.Minor_version and value is registry
            itself.
        """
        return LazyRegistries(self)

    def get_task_monitor(self, task_monitor_uri):
        """Used to retrieve a TaskMonitor by task monitor URI.

        :param task_monitor_uri: Task monitor URI
        :returns: A task monitor.
        """
        return taskmonitor.TaskMonitor(
            self._conn,
            task_monitor_uri,
            redfish_version=self.redfish_version,
            registries=self.registries)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Redfish standard schema.
# https://redfish.dmtf.org/schemas/v1/EventService.v1_0_8.json

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.eventservice import constants
from sushy.resources.eventservice import eventdestination

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):

    submit_test_event = common.ActionField(
        '#EventService.SubmitTestEvent')


class EventService(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The EventService resource identity"""

    name = base.Field('Name', required=True)
    """The EventService resource name"""

    delivery_retry_attempts = base.Field('DeliveryRetryAttempts')
    """Number of attempts an event posting is retried before the subscription
    is terminated. This retry is at the service level, meaning the HTTP POST
    to the Event Destination was returned by the HTTP operation as unsuccessful
    (4xx or 5xx return code) or an HTTP timeout occurred this many times before
    the Event Destination subscription is terminated
    """

    delivery_retry_interval = base.Field('DeliveryRetryIntervalSeconds')
    """Number of seconds between retry attempts for sending any given Event"""

    event_types_for_subscription = base.Field('EventTypesForSubscription',
                                              adapter=list)
    """Types of Events that can be subscribed to"""

    service_enabled = base.Field('ServiceEnabled', adapter=bool)
    """Indicates whether the EventService is enabled"""

    status = common.StatusField('Status')
    """The status of the EventService"""

    _actions = ActionsField('Actions', required=True)

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, root=None):
        """A class representing a EventService

        :param connector: A Connector instance
        :param identity: The identity of the EventService resource
        :param redfish_version: The version of Redfish. Used to construct
            the object according to schema of given version
        :param registries: Dict of registries to be used in any resource
            that needs registries to parse messages.
        :param root: Sushy root object. Empty for Sushy root itself.
        """
        super().__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries, root=root)

    def _get_submit_test_event(self):
        submit_test_event_action = self._actions.submit_test_event
        if not submit_test_event_action:
            raise exceptions.MissingActionError(
                action='#EventService.SubmitTestEvent',
                resource=self._path)
        return submit_test_event_action

    def submit_test_event(self, event_id, event_timestamp, event_type,
                          message, message_args, message_id, origin,
                          severity):
        """Submit Test Event is used to to send a test event to the BMC

        :param event_id: ID of event to be added.
        :param event_timestamp: time stamp of event to be added.
        :param event_type: type of event to be added.
        :param message: human readable message of event to be added.
        :param message_args: array of message arguments of the event to
            be added.
        :param message_id: message ID of event to be added.
        :param origin: string of the URL within the OriginOfCondition
            property of the event to be added
        :param severity: the Severity of event to be added.
        :param target: The  link to invoke action.
        :raises: MissingActionError if the EvenService does not have the
             action.
        """

        target_uri = self._get_submit_test_event().target_uri

        data = {
            'EventId': event_id,
            'EventTimestamp': event_timestamp,
            'EventType': event_type,
            'Message': message,
            'MessageArgs': message_args,
            'MessageId': message_id,
            'OriginOfCondition': origin,
            'Severity': severity
        }

        self._conn.post(target_uri, data=data)

    def get_event_types_for_subscription(self):
        """Get the Types of Events that can be subscribed to

        :returns: A set with the types of Events that can be subscribed to.
        """
        if not self.event_types_for_subscription:
            LOG.warning('Could not figure out the Event types supported by '
                        'the EventService %s', self.identity)
            return set(constants.EventType)

        return {v for v in constants.EventType
                if v.value in self.event_types_for_subscription}

    def _get_subscriptions_collection_path(self):
        """Helper function to find the EventDestinationCollections path"""
        subscriptions = self.json.get('Subscriptions')
        if not subscriptions:
            raise exceptions.MissingAttributeError(
                attribute='Subscriptions', resource=self._path)
        return subscriptions.get('@odata.id')

    @property
    def subscriptions(self):
        """Reference to a collection of Event Destination resources"""
        return eventdestination.EventDestinationCollection(
            self._conn, self._get_subscriptions_collection_path(),
            redfish_version=self.redfish_version, registries=self.registries,
            root=self.root)




# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from unittest import mock


import sushy
from sushy import exceptions
from sushy.resources.chassis import chassis
from sushy.resources.manager import manager
from sushy.resources.system.network import adapter
from sushy.resources.system import system
from sushy.tests.unit import base


class ChassisTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/chassis.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc
        self.conn.get.return_value.headers = {'ETag': 'd37f7bcd528e4d59'}

        self.chassis = chassis.Chassis(self.conn, '/redfish/v1/Chassis/Blade1',
                                       redfish_version='1.8.0')

    def test__parse_attributes(self):
        # | WHEN |
        self.chassis._parse_attributes(self.json_doc)
        # | THEN |
        self.assertEqual('1.8.0', self.chassis.redfish_version)
        self.assertEqual('Blade1', self.chassis.identity)
        self.assertEqual('Blade', self.chassis.name)
        self.assertEqual('Test description', self.chassis.description)
        self.assertEqual('45Z-2381', self.chassis.asset_tag)
        self.assertEqual(sushy.ChassisType.BLADE, self.chassis.chassis_type)
        self.assertEqual('Contoso', self.chassis.manufacturer)
        self.assertEqual('SX1000', self.chassis.model)
        self.assertEqual('529QB9450R6', self.chassis.serial_number)
        self.assertEqual('6914260', self.chassis.sku)
        self.assertEqual('166480-S23', self.chassis.part_number)
        self.assertEqual('FFFFFFFF-FFFF-FFFF-FFFF-FFFFFFFFFFFF',
                         self.chassis.uuid)
        self.assertEqual(sushy.IndicatorLED.OFF,
                         self.chassis.indicator_led)
        self.assertEqual(sushy.PowerState.ON, self.chassis.power_state)
        self.assertEqual(sushy.State.ENABLED, self.chassis.status.state)
        self.assertEqual(44.45, self.chassis.height_mm)
        self.assertEqual(431.8, self.chassis.width_mm)
        self.assertEqual(711, self.chassis.depth_mm)
        self.assertEqual(15.31, self.chassis.weight_kg)
        self.assertEqual(sushy.Health.OK, self.chassis.status.health)
        self.assertEqual(sushy.IntrusionSensor.NORMAL,
                         self.chassis.physical_security.intrusion_sensor)
        self.assertEqual(123,
                         self.chassis.physical_security.intrusion_sensor_number
                         )
        self.assertEqual(sushy.IntrusionSensorReArm.MANUAL,
                         self.chassis.physical_security.intrusion_sensor_re_arm
                         )

    def test__parse_attributes_return(self):
        attributes = self.chassis._parse_attributes(self.json_doc)

        # Test that various types are returned correctly
        self.assertEqual('Blade', attributes.get('name'))
        self.assertEqual(sushy.IndicatorLED.OFF,
                         attributes.get('indicator_led'))
        self.assertEqual(sushy.PowerState.ON, attributes.get('power_state'))
        self.assertEqual({'intrusion_sensor': sushy.IntrusionSensor.NORMAL,
                          'intrusion_sensor_number': 123,
                          'intrusion_sensor_re_arm':
                          sushy.IntrusionSensorReArm.MANUAL},
                         attributes.get('physical_security'))

    def test_get_allowed_reset_chasis_values(self):
        # | GIVEN |
        expected = {sushy.ResetType.POWER_CYCLE,
                    sushy.ResetType.PUSH_POWER_BUTTON,
                    sushy.ResetType.FORCE_ON, sushy.ResetType.NMI,
                    sushy.ResetType.FORCE_RESTART,
                    sushy.ResetType.GRACEFUL_RESTART, sushy.ResetType.ON,
                    sushy.ResetType.FORCE_OFF,
                    sushy.ResetType.GRACEFUL_SHUTDOWN}
        # | WHEN |
        values = self.chassis.get_allowed_reset_chassis_values()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_allowed_reset_chassis_values_for_no_values_set(self):
        # | GIVEN |
        self.chassis._actions.reset.allowed_values = []
        expected = {sushy.ResetType.POWER_CYCLE,
                    sushy.ResetType.PUSH_POWER_BUTTON,
                    sushy.ResetType.FORCE_ON, sushy.ResetType.NMI,
                    sushy.ResetType.FORCE_RESTART,
                    sushy.ResetType.GRACEFUL_RESTART, sushy.ResetType.ON,
                    sushy.ResetType.FORCE_OFF,
                    sushy.ResetType.GRACEFUL_SHUTDOWN,
                    sushy.ResetType.SUSPEND,
                    sushy.ResetType.RESUME,
                    sushy.ResetType.PAUSE}
        # | WHEN |
        values = self.chassis.get_allowed_reset_chassis_values()
        # | THEN |
        self.assertEqual(expected, values)
        self.assertIsInstance(values, set)

    def test_get_allowed_reset_chassis_values_missing_action_reset_attr(self):
        # | GIVEN |
        self.chassis._actions.reset = None
        # | WHEN & THEN |
        self.assertRaisesRegex(
            exceptions.MissingActionError, 'action #Chassis.Reset')

    def test_reset_chassis(self):
        self.chassis.reset_chassis(sushy.ResetType.GRACEFUL_RESTART)
        self.chassis._conn.post.assert_called_once_with(
            '/redfish/v1/Chassis/Blade1/Actions/Chassis.Reset',
            data={'ResetType': 'GracefulRestart'})

    def test_reset_chassis_with_invalid_value(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.chassis.reset_chassis, 'invalid-value')

    def test_set_indicator_led(self):
        with mock.patch.object(
                self.chassis, 'invalidate', autospec=True) as invalidate_mock:
            self.chassis.set_indicator_led(sushy.IndicatorLED.BLINKING)
            self.chassis._conn.patch.assert_called_once_with(
                '/redfish/v1/Chassis/Blade1',
                data={'IndicatorLED': 'Blinking'},
                etag='d37f7bcd528e4d59')

            invalidate_mock.assert_called_once_with()

    def test_set_indicator_led_invalid_state(self):
        self.assertRaises(exceptions.InvalidParameterValueError,
                          self.chassis.set_indicator_led,
                          'spooky-glowing')

    def test_managers(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'manager.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        actual_managers = self.chassis.managers
        self.assertIsInstance(actual_managers[0], manager.Manager)
        self.assertEqual(
            '/redfish/v1/Managers/Blade1BMC', actual_managers[0].path)

    def test_systems(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/'
                  'system.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        # | WHEN & THEN |
        actual_systems = self.chassis.systems
        self.assertIsInstance(actual_systems[0], system.System)
        self.assertEqual(
            '/redfish/v1/Systems/529QB9450R6', actual_systems[0].path)

    def test_network_adapters(self):
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/'
                  'json_samples/network_adapter_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        actual_adapters = self.chassis.network_adapters
        self.assertIsInstance(actual_adapters,
                              adapter.NetworkAdapterCollection)
        self.conn.get.return_value.json.assert_called_once_with()

    def test_network_adapters_cached(self):
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/'
                  'json_samples/network_adapter_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        actual_adapters = self.chassis.network_adapters
        self.conn.get.return_value.json.reset_mock()
        self.assertIs(actual_adapters,
                      self.chassis.network_adapters)
        self.conn.get.return_value.json.assert_not_called()

    def test_network_adapters_refresh(self):
        with open('sushy/tests/unit/'
                  'json_samples/network_adapter_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        adapters = self.chassis.network_adapters
        self.assertIsInstance(adapters, adapter.NetworkAdapterCollection)

        with open('sushy/tests/unit/'
                  'json_samples/chassis.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.chassis.invalidate()
        self.chassis.refresh(force=False)

        self.assertTrue(adapters._is_stale)

        with open('sushy/tests/unit/'
                  'json_samples/network_adapter_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.assertIsInstance(self.chassis.network_adapters,
                              adapter.NetworkAdapterCollection)

    def test_get_expanded_data_with_full_data(self):
        # Test that expanded data is returned when present
        expanded_data = {'Key': [{'Name': 'Val'}], 'List': []}
        self.chassis._json = {'Power': expanded_data}
        result = self.chassis._get_expanded_data('Power')
        self.assertEqual(expanded_data, result)

    def test_get_expanded_data_with_reference_only(self):
        # Test that None is returned for reference-only data
        reference_data = {'@odata.id': '/redfish/v1/Chassis/1/Power'}
        self.chassis._json = {'Power': reference_data}
        result = self.chassis._get_expanded_data('Power')
        self.assertIsNone(result)

    def test_get_expanded_data_missing_field(self):
        # Test that None is returned when field is missing
        self.chassis._json = {}
        result = self.chassis._get_expanded_data('Power')
        self.assertIsNone(result)


class ChassisCollectionTestCase(base.TestCase):

    def setUp(self):
        super().setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'chassis_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.chassis = chassis.ChassisCollection(
            self.conn, '/redfish/v1/Chassis', redfish_version='1.5.0')

    @mock.patch.object(chassis, 'Chassis', autospec=True)
    def test_get_member(self, chassis_mock):
        self.chassis.get_member('/redfish/v1/Chassis/MultiBladeEncl')
        chassis_mock.assert_called_once_with(
            self.chassis._conn, '/redfish/v1/Chassis/MultiBladeEncl',
            redfish_version=self.chassis.redfish_version, registries=None,
            root=self.chassis.root)

    @mock.patch.object(chassis, 'Chassis', autospec=True)
    def test_get_members(self, chassis_mock):
        members = self.chassis.get_members()
        calls = [
            mock.call(self.chassis._conn, '/redfish/v1/Chassis/MultiBladeEncl',
                      redfish_version=self.chassis.redfish_version,
                      registries=None,
                      root=self.chassis.root),
            mock.call(self.chassis._conn, '/redfish/v1/Chassis/Blade1',
                      redfish_version=self.chassis.redfish_version,
                      registries=None,
                      root=self.chassis.root),
            mock.call(self.chassis._conn, '/redfish/v1/Chassis/Blade2',
                      redfish_version=self.chassis.redfish_version,
                      registries=None,
                      root=self.chassis.root),
            mock.call(self.chassis._conn, '/redfish/v1/Chassis/Blade3',
                      redfish_version=self.chassis.redfish_version,
                      registries=None,
                      root=self.chassis.root),
            mock.call(self.chassis._conn, '/redfish/v1/Chassis/Blade4',
                      redfish_version=self.chassis.redfish_version,
                      registries=None,
                      root=self.chassis.root)
        ]
        chassis_mock.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(5, len(members))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values come from the Redfish UpdateService json-schema.
# https://redfish.dmtf.org/schemas/UpdateService.v1_2_2.json#/definitions/TransferProtocolType

import enum


class UpdateTransferProtocolType(enum.Enum):
    """Transfer Protocol Type constants"""

    CIFS = 'CIFS'
    """Common Internet File System (CIFS)."""

    FTP = 'FTP'
    """File Transfer Protocol (FTP)."""

    SFTP = 'SFTP'
    """Secure File Transfer Protocol (SFTP)."""

    HTTP = 'HTTP'
    """Hypertext Transfer Protocol (HTTP)."""

    HTTPS = 'HTTPS'
    """Hypertext Transfer Protocol Secure (HTTPS)."""

    SCP = 'SCP'
    """Secure Copy Protocol (SCP)."""

    TFTP = 'TFTP'
    """Trivial File Transfer Protocol (TFTP)."""

    OEM = 'OEM'
    """A manufacturer-defined protocol."""

    NFS = 'NFS'
    """Network File System (NFS)."""

    # Deprecated alias:

    NSF = 'NFS'
    """Network File System (NFS)."""


UPDATE_PROTOCOL_CIFS = UpdateTransferProtocolType.CIFS
UPDATE_PROTOCOL_FTP = UpdateTransferProtocolType.FTP
UPDATE_PROTOCOL_SFTP = UpdateTransferProtocolType.SFTP
UPDATE_PROTOCOL_HTTP = UpdateTransferProtocolType.HTTP
UPDATE_PROTOCOL_HTTPS = UpdateTransferProtocolType.HTTPS
UPDATE_PROTOCOL_SCP = UpdateTransferProtocolType.SCP
UPDATE_PROTOCOL_TFTP = UpdateTransferProtocolType.TFTP
UPDATE_PROTOCOL_OEM = UpdateTransferProtocolType.OEM
UPDATE_PROTOCOL_NFS = UpdateTransferProtocolType.NFS
